from __future__ import annotations

import contextlib
import functools
import importlib.util
import logging
from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any, Generic, Optional, cast

from typing_extensions import (
    ParamSpec,
    Protocol,
    TypeAliasType,
    TypeVar,
    override,
    runtime_checkable,
)

from .utils import LovelyStats, format_tensor_stats

log = logging.getLogger(__name__)

TArray = TypeVar("TArray", infer_variance=True)
P = ParamSpec("P")

LovelyStatsFn = TypeAliasType(
    "LovelyStatsFn",
    Callable[[TArray], Optional[LovelyStats]],
    type_params=(TArray,),
)


@runtime_checkable
class LovelyReprFn(Protocol[TArray]):
    @property
    def __lovely_repr_instance__(self) -> lovely_repr[TArray]: ...

    @__lovely_repr_instance__.setter
    def __lovely_repr_instance__(self, value: lovely_repr[TArray]) -> None: ...

    @property
    def __name__(self) -> str: ...

    def set_fallback_repr(self, repr_fn: Callable[[TArray], str]) -> None: ...
    def __call__(self, value: TArray, /) -> str: ...


def _find_missing_deps(dependencies: list[str]):
    missing_deps: list[str] = []

    for dep in dependencies:
        if importlib.util.find_spec(dep) is not None:
            continue

        missing_deps.append(dep)

    return missing_deps


class lovely_repr(Generic[TArray]):
    @override
    def __init__(
        self,
        dependencies: list[str],
        fallback_repr: Callable[[TArray], str] | None = None,
    ):
        """
        Decorator to create a lovely representation function for an array.

        Args:
            dependencies: List of dependencies to check before running the function.
                If any dependency is not available, the function will not run.
            fallback_repr: A function that takes an array and returns its fallback representation.
        Returns:
            A decorator function that takes a function and returns a lovely representation function.

        Example:
            @lovely_repr(dependencies=["torch"])
            def my_array_stats(array):
                return {...}
        """
        super().__init__()

        if fallback_repr is None:
            fallback_repr = repr

        self._dependencies = dependencies
        self._fallback_repr = fallback_repr

    def set_fallback_repr(self, repr_fn: Callable[[TArray], str]) -> None:
        self._fallback_repr = repr_fn

    def __call__(
        self, array_stats_fn: LovelyStatsFn[TArray], /
    ) -> LovelyReprFn[TArray]:
        @functools.wraps(array_stats_fn)
        def wrapper_fn(array: TArray) -> str:
            if missing_deps := _find_missing_deps(self._dependencies):
                log.warning(
                    f"Missing dependencies: {', '.join(missing_deps)}. "
                    "Skipping lovely representation."
                )
                return self._fallback_repr(array)

            if (stats := array_stats_fn(array)) is None:
                return self._fallback_repr(array)

            return format_tensor_stats(stats)

        wrapper = cast(LovelyReprFn[TArray], wrapper_fn)
        wrapper.__lovely_repr_instance__ = self
        wrapper.set_fallback_repr = self.set_fallback_repr
        return wrapper


class lovely_patch(contextlib.AbstractContextManager["lovely_patch"], ABC):
    def __init__(self):
        self._patched = False
        self.__enter__()

    def dependencies(self) -> list[str]:
        """Subclasses can override this to specify the dependencies of the patch."""
        return []

    @abstractmethod
    def patch(self):
        """Subclasses must implement this."""

    @abstractmethod
    def unpatch(self):
        """Subclasses must implement this."""

    @override
    def __enter__(self):
        if self._patched:
            return self

        if missing_deps := _find_missing_deps(self.dependencies()):
            log.warning(
                f"Missing dependencies: {', '.join(missing_deps)}. "
                "Skipping monkey patch."
            )
            return self

        self.patch()
        self._patched = True
        return self

    @override
    def __exit__(self, *exc_info):
        if not self._patched:
            return

        self.unpatch()
        self._patched = False

    def close(self):
        """Explicitly clean up the resource."""
        self.__exit__(None, None, None)



from __future__ import annotations

import contextlib
import importlib.util
import logging
from typing import Literal

from typing_extensions import TypeAliasType, assert_never, override

from ._base import lovely_patch

Library = TypeAliasType("Library", Literal["numpy", "torch", "jax"])

log = logging.getLogger(__name__)


def _find_deps() -> list[Library]:
    """
    Find available libraries for monkey patching.
    """
    deps: list[Library] = []
    if importlib.util.find_spec("torch") is not None:
        deps.append("torch")
    if importlib.util.find_spec("jax") is not None:
        deps.append("jax")
    if importlib.util.find_spec("numpy") is not None:
        deps.append("numpy")
    return deps


class monkey_patch(lovely_patch):
    def __init__(self, libraries: list[Library] | Literal["auto"] = "auto"):
        if libraries == "auto":
            self.libraries = _find_deps()
        else:
            self.libraries = libraries

        if not self.libraries:
            raise ValueError(
                "No libraries found for monkey patching. "
                "Please install numpy, torch, or jax."
            )

        self.stack = contextlib.ExitStack()
        super().__init__()

    @override
    def patch(self):
        for library in self.libraries:
            if library == "torch":
                from .torch_ import torch_monkey_patch

                self.stack.enter_context(torch_monkey_patch())
            elif library == "jax":
                from .jax_ import jax_monkey_patch

                self.stack.enter_context(jax_monkey_patch())
            elif library == "numpy":
                from .numpy_ import numpy_monkey_patch

                self.stack.enter_context(numpy_monkey_patch())
            else:
                assert_never(library)

        log.info(
            f"Monkey patched libraries: {', '.join(self.libraries)}. "
            "You can now use the lovely functions with these libraries."
        )

    @override
    def unpatch(self):
        self.stack.close()
        log.info(
            f"Unmonkey patched libraries: {', '.join(self.libraries)}. "
            "You can now use the lovely functions with these libraries."
        )



from __future__ import annotations

from ._assert import assert_ as assert_
from ._config import ROOT_PATH as ROOT_PATH
from ._config import DebugConfig as DebugConfig
from ._config import config as config
from ._config import enabled as enabled
from ._config import override as override
from ._decorators import ensure as ensure
from ._decorators import ensure as post
from ._decorators import require as pre
from ._decorators import require as require
from ._decorators import snapshot as snapshot
from ._globals import aRepr as aRepr
from ._lightning import SanityDebugFlagCallback as SanityDebugFlagCallback
from ._metaclass import DBC as DBC
from ._metaclass import DBCMeta as DBCMeta
from ._types import Contract as Contract
from ._types import InvariantCheckEvent as InvariantCheckEvent
from ._types import Snapshot as Snapshot
from .errors import ViolationError as ViolationError

_, _ = pre, post
# from ._decorators import invariant as invariant # not supported yet



from __future__ import annotations

import contextlib
import importlib.metadata
import warnings
from collections.abc import Callable
from typing import Any, Protocol, cast

from typing_extensions import TypeVar

T = TypeVar("T", infer_variance=True)
TFn = TypeVar("TFn", bound=Callable)


class _SnoopInstance(Protocol):
    def __enter__(self) -> None: ...

    def __exit__(self, exc_type, exc_value, traceback) -> None: ...

    def __call__(self, fn: TFn) -> TFn: ...


class SnoopConstructor(Protocol):
    def __call__(self, *args, **kwargs) -> _SnoopInstance: ...

    def disable(self) -> _SnoopInstance: ...


snoop: SnoopConstructor
try:
    import warnings
    from contextlib import nullcontext

    import pysnooper  # pyright: ignore[reportMissingImports]
    import pysnooper.utils  # pyright: ignore[reportMissingImports]

    try:
        import torch  # pyright: ignore[reportMissingImports]
    except ImportError:
        torch = None

    try:
        import numpy  # pyright: ignore[reportMissingImports]
    except ImportError:
        numpy = None

    try:
        import jax  # pyright: ignore[reportMissingImports]
    except ImportError:
        jax = None

    FLOATING_POINTS = set()
    for i in ["float", "double", "half", "complex128", "complex32", "complex64"]:
        # older version of PyTorch do not have complex dtypes
        if torch is None or not hasattr(torch, i):
            continue
        FLOATING_POINTS.add(getattr(torch, i))

    try:
        __version__ = importlib.metadata.version(__name__)
    except importlib.metadata.PackageNotFoundError:
        # package is not installed
        pass

    def default_torch_format(x):
        try:
            from .lovely import torch_repr

            return torch_repr(x)
        except BaseException:
            return str(x.shape)

    def default_numpy_format(x):
        try:
            from .lovely import numpy_repr

            return numpy_repr(x)
        except BaseException:
            return str(x.shape)

    def default_jax_format(x):
        try:
            from .lovely import jax_repr

            return jax_repr(x)
        except BaseException:
            return str(x.shape)

    class TorchSnooper(pysnooper.tracer.Tracer):
        def __init__(
            self,
            *args,
            tensor_format=default_torch_format,
            numpy_format=default_numpy_format,
            jax_format=default_jax_format,
            **kwargs,
        ):
            self.orig_custom_repr = (
                kwargs["custom_repr"] if "custom_repr" in kwargs else ()
            )
            custom_repr = (lambda x: True, self.compute_repr)
            kwargs["custom_repr"] = (custom_repr,)
            super(TorchSnooper, self).__init__(*args, **kwargs)
            self.tensor_format = tensor_format
            self.numpy_format = numpy_format
            self.jax_format = jax_format

        @staticmethod
        def is_return_types(x):
            return type(x).__module__ == "torch.return_types"

        def return_types_repr(self, x):
            if type(x).__name__ in {
                "max",
                "min",
                "median",
                "mode",
                "sort",
                "topk",
                "kthvalue",
            }:
                return (
                    type(x).__name__
                    + "(values="
                    + self.tensor_format(x.values)
                    + ", indices="
                    + self.tensor_format(x.indices)
                    + ")"
                )
            if type(x).__name__ == "svd":
                return (
                    "svd(U="
                    + self.tensor_format(x.U)
                    + ", S="
                    + self.tensor_format(x.S)
                    + ", V="
                    + self.tensor_format(x.V)
                    + ")"
                )
            if type(x).__name__ == "slogdet":
                return (
                    "slogdet(sign="
                    + self.tensor_format(x.sign)
                    + ", logabsdet="
                    + self.tensor_format(x.logabsdet)
                    + ")"
                )
            if type(x).__name__ == "qr":
                return (
                    "qr(Q="
                    + self.tensor_format(x.Q)
                    + ", R="
                    + self.tensor_format(x.R)
                    + ")"
                )
            if type(x).__name__ == "solve":
                return (
                    "solve(solution="
                    + self.tensor_format(x.solution)
                    + ", LU="
                    + self.tensor_format(x.LU)
                    + ")"
                )
            if type(x).__name__ == "geqrf":
                return (
                    "geqrf(a="
                    + self.tensor_format(x.a)
                    + ", tau="
                    + self.tensor_format(x.tau)
                    + ")"
                )
            if type(x).__name__ in {"symeig", "eig"}:
                return (
                    type(x).__name__
                    + "(eigenvalues="
                    + self.tensor_format(x.eigenvalues)
                    + ", eigenvectors="
                    + self.tensor_format(x.eigenvectors)
                    + ")"
                )
            if type(x).__name__ == "triangular_solve":
                return (
                    "triangular_solve(solution="
                    + self.tensor_format(x.solution)
                    + ", cloned_coefficient="
                    + self.tensor_format(x.cloned_coefficient)
                    + ")"
                )
            if type(x).__name__ == "gels":
                return (
                    "gels(solution="
                    + self.tensor_format(x.solution)
                    + ", QR="
                    + self.tensor_format(x.QR)
                    + ")"
                )
            raise ValueError("Unknown return_types encountered, open a bug report!")

        def compute_repr(self, x):
            orig_repr_func = pysnooper.utils.get_repr_function(x, self.orig_custom_repr)
            if torch is not None and torch.is_tensor(x):
                return self.tensor_format(x)
            if numpy is not None and isinstance(x, numpy.ndarray):
                return self.numpy_format(x)
            if jax is not None and isinstance(x, jax.Array):
                return self.jax_format(x)
            if self.is_return_types(x):
                return self.return_types_repr(x)
            if orig_repr_func is not repr:
                return orig_repr_func(x)
            if isinstance(x, (list, tuple)):
                content = ""
                for i in x:
                    if content != "":
                        content += ", "
                    content += self.compute_repr(i)
                if isinstance(x, tuple) and len(x) == 1:
                    content += ","
                if isinstance(x, tuple):
                    return "(" + content + ")"
                return "[" + content + "]"
            if isinstance(x, dict):
                content = ""
                for k, v in x.items():
                    if content != "":
                        content += ", "
                    content += self.compute_repr(k) + ": " + self.compute_repr(v)
                return "{" + content + "}"
            return repr(x)

    class _Snoop:
        disable = nullcontext
        __call__ = TorchSnooper

    snoop = cast(Any, _Snoop())

except ImportError:
    from typing_extensions import override

    class _snoop_cls(contextlib.nullcontext, contextlib.ContextDecorator):
        def __init__(self, *, _should_warn: bool = True):
            super().__init__()

            self._should_warn = _should_warn
            self._has_warned = False

        @classmethod
        def disable(cls):
            return cls(_should_warn=False)

        @override
        def __enter__(self):
            if self._should_warn and not self._has_warned:
                warnings.warn(
                    "snoop is not installed, please install it to enable snoop"
                )
                self._has_warned = True

            return super().__enter__()

    snoop = cast(Any, _snoop_cls)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import numpy as np
from typing_extensions import override

from ._base import lovely_patch, lovely_repr
from .utils import LovelyStats, array_stats, patch_to

if TYPE_CHECKING:
    import jax  # pyright: ignore[reportMissingImports]


def _type_name(array: jax.Array):
    type_name = type(array).__name__.rsplit(".", 1)[-1]
    return "array" if type_name == "ArrayImpl" else type_name


_DT_NAMES = {
    "float16": "f16",
    "float32": "f32",
    "float64": "f64",
    "uint8": "u8",
    "uint16": "u16",
    "uint32": "u32",
    "uint64": "u64",
    "int8": "i8",
    "int16": "i16",
    "int32": "i32",
    "int64": "i64",
    "bfloat16": "bf16",
    "complex64": "c64",
    "complex128": "c128",
}


def _dtype_str(array: jax.Array) -> str:
    dtype_base = str(array.dtype).rsplit(".", 1)[-1]
    dtype_base = _DT_NAMES.get(dtype_base, dtype_base)
    return dtype_base


def _device(array: jax.Array) -> str:
    from jaxlib.xla_extension import Device  # pyright: ignore[reportMissingImports]

    if callable(device := array.device):
        device = device()

    device = cast(Device, device)
    if device.platform == "cpu":
        return "cpu"

    return f"{device.platform}:{device.id}"


@lovely_repr(dependencies=["jax"])
def jax_repr(array: jax.Array) -> LovelyStats | None:
    import jax.numpy as jnp  # pyright: ignore[reportMissingImports]

    # For dtypes like `object` or `str`, we let the fallback repr handle it
    if not jnp.issubdtype(array.dtype, jnp.number):
        return None

    return {
        # Basic attributes
        "shape": array.shape,
        "size": array.size,
        "nbytes": array.nbytes,
        "type_name": _type_name(array),
        # Dtype
        "dtype_str": _dtype_str(array),
        "is_complex": jnp.iscomplexobj(array),
        # Device
        "device": _device(array),
        # Depending of whether the tensor is complex or not, we will call the appropriate stats function
        **array_stats(np.asarray(array)),
    }


class jax_monkey_patch(lovely_patch):
    @override
    def dependencies(self) -> list[str]:
        return ["jax"]

    @override
    def patch(self):
        from jax._src import array  # pyright: ignore[reportMissingImports]

        self.prev_repr = array.ArrayImpl.__repr__
        self.prev_str = array.ArrayImpl.__str__
        jax_repr.set_fallback_repr(self.prev_repr)

        patch_to(array.ArrayImpl, "__repr__", jax_repr)
        patch_to(array.ArrayImpl, "__str__", jax_repr)

    @override
    def unpatch(self):
        from jax._src import array  # pyright: ignore[reportMissingImports]

        patch_to(array.ArrayImpl, "__repr__", self.prev_repr)
        patch_to(array.ArrayImpl, "__str__", self.prev_str)



from __future__ import annotations

import importlib.util
from functools import cache
from typing import Any


@cache
def _in_ipython():
    try:
        from IPython import get_ipython

        return get_ipython() is not None
    except ImportError:
        return False


@cache
def _treescope_installed():
    return importlib.util.find_spec("treescope") is not None


@cache
def _rich_installed():
    return importlib.util.find_spec("rich") is not None


def display(*args: Any):
    """
    Display the given arguments in the current environment.

    If executed in an IPython environment, the display will be handled
    by treescope if installed, or rich if available. If neither are
    installed, it will fall back to IPython's display function. In a
    non-IPython environment, rich will be used if available, otherwise
    the standard print function will be used.

    Args:
        *args: Any objects to display.
    """
    if _in_ipython():
        if _treescope_installed():
            import treescope

            with treescope.active_autovisualizer.set_scoped(
                treescope.ArrayAutovisualizer()
            ):
                treescope.display(*args)
        elif _rich_installed():
            import rich

            rich.print(*args)
        else:
            from IPython.display import display

            display(*args)
    elif _rich_installed():
        import rich

        rich.print(*args)
    else:
        print(*args)



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .lovely._monkey_patch_all import Library


def setup_logging(
    *,
    lovely: bool | list[Library] = False,
    treescope: bool = False,
    treescope_autovisualize_arrays: bool = False,
    rich: bool = False,
    rich_tracebacks: bool = False,
    log_level: int | str | None = logging.INFO,
    log_save_dir: Path | None = None,
):
    if lovely:
        from .lovely._monkey_patch_all import monkey_patch

        monkey_patch("auto" if lovely is True else lovely)

    if treescope:
        try:
            # Check if we're in a Jupyter environment
            from IPython import get_ipython

            if get_ipython() is not None:
                import treescope as _treescope  # type: ignore

                _treescope.basic_interactive_setup(
                    autovisualize_arrays=treescope_autovisualize_arrays
                )
            else:
                logging.info(
                    "Treescope setup is only supported in Jupyter notebooks. Skipping."
                )
        except ImportError:
            logging.info(
                "Failed to import `treescope` or `IPython`. Ignoring `treescope` registration"
            )

    log_handlers: list[logging.Handler] = []
    if log_save_dir:
        log_file = log_save_dir / "logging.log"
        log_file.touch(exist_ok=True)
        log_handlers.append(logging.FileHandler(log_file))

    if rich:
        try:
            from rich.logging import RichHandler  # type: ignore

            log_handlers.append(RichHandler(rich_tracebacks=rich_tracebacks))
        except ImportError:
            logging.info(
                "Failed to import rich. Falling back to default Python logging."
            )

    logging.basicConfig(
        level=log_level,
        format="%(message)s",
        datefmt="[%X]",
        handlers=log_handlers,
    )
    logging.info(
        "Logging initialized. "
        f"Lovely: {lovely}, Treescope: {treescope}, Rich: {rich}, "
        f"Log level: {log_level}, Log save dir: {log_save_dir}"
    )


init_python_logging = setup_logging



from __future__ import annotations

from dataclasses import dataclass
from typing import ClassVar

from typing_extensions import Self


@dataclass
class LovelyConfig:
    """
    This class is used to manage the configuration of the Lovely library.
    """

    precision: int = 3
    """Number of digits after the decimal point."""

    threshold_max: int = 3
    """Absolute values larger than 10^3 use scientific notation."""

    threshold_min: int = -4
    """Absolute values smaller than 10^-4 use scientific notation."""

    sci_mode: bool | None = None
    """Force scientific notation (None=auto)."""

    show_mem_above: int = 1024
    """Show memory size if above this threshold (bytes)."""

    color: bool = True
    """Use ANSI colors in text."""

    indent: int = 2
    """Indent for nested representation."""

    config_instance: ClassVar[Self | None] = None
    """Singleton instance of the LovelyConfig class."""

    @classmethod
    def instance(cls) -> Self:
        """
        Get the singleton instance of the LovelyConfig class.
        If it doesn't exist, create it.
        """
        if cls.config_instance is None:
            cls.config_instance = cls()
        return cls.config_instance



from __future__ import annotations

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach_stub(__name__, __file__)


try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:
    # For Python <3.8
    from importlib_metadata import (  # pyright: ignore[reportMissingImports]
        PackageNotFoundError,
        version,
    )

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    __version__ = "unknown"



from __future__ import annotations

import logging

import numpy as np
from typing_extensions import override

from ._base import lovely_patch, lovely_repr
from .utils import LovelyStats, array_stats


def _np_ge_2():
    import importlib.metadata

    from packaging.version import Version

    try:
        numpy_version = importlib.metadata.version("numpy")
        return Version(numpy_version) >= Version("2.0")
    except importlib.metadata.PackageNotFoundError:
        return False


def _type_name(array: np.ndarray):
    return (
        "array"
        if type(array) is np.ndarray
        else type(array).__name__.rsplit(".", 1)[-1]
    )


_DT_NAMES = {
    "float16": "f16",
    "float32": "f32",
    "float64": "",  # Default dtype in numpy
    "uint8": "u8",
    "uint16": "u16",
    "uint32": "u32",
    "uint64": "u64",
    "int8": "i8",
    "int16": "i16",
    "int32": "i32",
    "int64": "i64",
    "complex64": "c64",
    "complex128": "c128",
}


def _dtype_str(array: np.ndarray) -> str:
    dtype_base = str(array.dtype).rsplit(".", 1)[-1]
    dtype_base = _DT_NAMES.get(dtype_base, dtype_base)
    return dtype_base


@lovely_repr(dependencies=["numpy"])
def numpy_repr(array: np.ndarray) -> LovelyStats | None:
    # For dtypes like `object` or `str`, we let the fallback repr handle it
    if not np.issubdtype(array.dtype, np.number):
        return None

    return {
        # Basic attributes
        "shape": array.shape,
        "size": array.size,
        "nbytes": array.nbytes,
        "type_name": _type_name(array),
        # Dtype
        "dtype_str": _dtype_str(array),
        "is_complex": np.iscomplexobj(array),
        # Depending of whether the tensor is complex or not, we will call the appropriate stats function
        **array_stats(array),
    }


numpy_repr.set_fallback_repr(np.array_repr)


class numpy_monkey_patch(lovely_patch):
    @override
    def dependencies(self) -> list[str]:
        return ["numpy"]

    @override
    def patch(self):
        if _np_ge_2():
            self.original_options = np.get_printoptions()
            np.set_printoptions(override_repr=numpy_repr)
            logging.info(
                f"Numpy monkey patching: using {numpy_repr.__name__} for numpy arrays. "
                f"{np.get_printoptions()=}"
            )
        else:
            # For legacy numpy, `set_string_function(None)` reverts to the default,
            # so no state needs to be saved.
            np.set_string_function(numpy_repr, True)  # pyright: ignore[reportAttributeAccessIssue]
            np.set_string_function(numpy_repr, False)  # pyright: ignore[reportAttributeAccessIssue]
            logging.info(
                f"Numpy monkey patching: using {numpy_repr.__name__} for numpy arrays. "
                f"{np.get_printoptions()=}"
            )

    @override
    def unpatch(self):
        if _np_ge_2():
            np.set_printoptions(**self.original_options)
            logging.info(
                f"Numpy unmonkey patching: using {numpy_repr.__name__} for numpy arrays. "
                f"{np.get_printoptions()=}"
            )
        else:
            np.set_string_function(None, True)  # pyright: ignore[reportAttributeAccessIssue]
            np.set_string_function(None, False)  # pyright: ignore[reportAttributeAccessIssue]
            logging.info(
                f"Numpy unmonkey patching: using {numpy_repr.__name__} for numpy arrays. "
                f"{np.get_printoptions()=}"
            )



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from typing_extensions import override

from ._base import lovely_patch, lovely_repr
from .utils import LovelyStats, array_stats, patch_to

if TYPE_CHECKING:
    import torch  # pyright: ignore[reportMissingImports]


def _type_name(tensor: torch.Tensor):
    import torch  # pyright: ignore[reportMissingImports]

    return (
        "tensor"
        if type(tensor) is torch.Tensor
        else type(tensor).__name__.split(".")[-1]
    )


_DT_NAMES = {
    "float32": "",  # Default dtype
    "float16": "f16",
    "float64": "f64",
    "bfloat16": "bf16",
    "uint8": "u8",
    "int8": "i8",
    "int16": "i16",
    "int32": "i32",
    "int64": "i64",
    "complex32": "c32",
    "complex64": "c64",
    "complex128": "c128",
}


def _dtype_str(tensor: torch.Tensor) -> str:
    dtype_base = str(tensor.dtype).rsplit(".", 1)[-1]
    dtype_base = _DT_NAMES.get(dtype_base, dtype_base)
    return dtype_base


def _to_np(tensor: torch.Tensor) -> np.ndarray:
    import torch  # pyright: ignore[reportMissingImports]

    # Get tensor data as CPU NumPy array for analysis
    t_cpu = tensor.detach().cpu()

    # Convert bfloat16 to float32 for numpy compatibility
    if tensor.dtype == torch.bfloat16:
        t_cpu = t_cpu.to(torch.float32)

    # Convert to NumPy
    t_np = t_cpu.numpy()

    return t_np


@lovely_repr(dependencies=["torch"])
def torch_repr(tensor: torch.Tensor) -> LovelyStats | None:
    return {
        # Basic attributes
        "shape": tensor.shape,
        "size": tensor.numel(),
        "nbytes": tensor.element_size() * tensor.numel(),
        "type_name": _type_name(tensor),
        # Device
        "device": str(tensor.device) if tensor.device else None,
        "is_meta": device.type == "meta" if (device := tensor.device) else False,
        # Grad
        "requires_grad": tensor.requires_grad,
        # Dtype
        "dtype_str": _dtype_str(tensor),
        "is_complex": tensor.is_complex(),
        # Depending of whether the tensor is complex or not, we will call the appropriate stats function
        **array_stats(_to_np(tensor)),
    }


class torch_monkey_patch(lovely_patch):
    @override
    def dependencies(self) -> list[str]:
        return ["torch"]

    @override
    def patch(self):
        import torch  # pyright: ignore[reportMissingImports]

        self.original_repr = torch.Tensor.__repr__
        self.original_str = torch.Tensor.__str__
        self.original_parameter_repr = torch.nn.Parameter.__repr__
        torch_repr.set_fallback_repr(self.original_repr)

        patch_to(torch.Tensor, "__repr__", torch_repr)
        patch_to(torch.Tensor, "__str__", torch_repr)
        try:
            delattr(torch.nn.Parameter, "__repr__")
        except AttributeError:
            pass

    @override
    def unpatch(self):
        import torch  # pyright: ignore[reportMissingImports]

        patch_to(torch.Tensor, "__repr__", self.original_repr)
        patch_to(torch.Tensor, "__str__", self.original_str)
        patch_to(torch.nn.Parameter, "__repr__", self.original_parameter_repr)



# Copyright (c) 2022 Google LLC
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import functools as ft
import typing
from typing import TYPE_CHECKING, Any, Generic, TypeVar

import wadler_lindig as wl
from jaxtyping._errors import AnnotationError
from jaxtyping._storage import (
    clear_treeflatten_memo,
    clear_treepath_memo,
    get_shape_memo,
    set_shape_memo,
    set_treeflatten_memo,
    set_treepath_memo,
)
from typing_extensions import override

from . import _tree_util as tree

_T = TypeVar("_T")
_S = TypeVar("_S")


class _FakePyTree1(Generic[_T]):
    pass


_FakePyTree1.__name__ = "PyTree"
_FakePyTree1.__qualname__ = "PyTree"
_FakePyTree1.__module__ = "builtins"


class _FakePyTree2(Generic[_T, _S]):
    pass


_FakePyTree2.__name__ = "PyTree"
_FakePyTree2.__qualname__ = "PyTree"
_FakePyTree2.__module__ = "builtins"


class _MetaPyTree(type):
    if TYPE_CHECKING:
        leaftype: Any
        structure: str | None

    @override
    def __call__(self, *args, **kwargs):
        raise RuntimeError("PyTree cannot be instantiated")

    @override
    def __instancecheck__(cls, obj):
        if not hasattr(cls, "leaftype"):
            return True  # Just `isinstance(x, PyTree)`
        # Handle beartype doing `isinstance(None, hint)` to check if
        # is `instance`able.
        if obj is None:
            return True

        single_memo, variadic_memo, pytree_memo, arg_memo = get_shape_memo()
        single_memo_bak = single_memo.copy()
        variadic_memo_bak = variadic_memo.copy()
        pytree_memo_bak = pytree_memo.copy()
        arg_memo_bak = arg_memo.copy()
        try:
            out = cls._check(obj, pytree_memo)
        except Exception:
            set_shape_memo(
                single_memo_bak, variadic_memo_bak, pytree_memo_bak, arg_memo_bak
            )
            raise
        if out:
            return True
        else:
            set_shape_memo(
                single_memo_bak, variadic_memo_bak, pytree_memo_bak, arg_memo_bak
            )
            return False

    def _check(cls, obj, pytree_memo):
        if cls.leaftype is Any:

            def is_flatten_leaftype(x) -> bool:
                return False

            def is_check_leaftype(x) -> bool:
                return True

        else:
            # We could use `isinstance` here but that would fail for more complicated
            # types, e.g. PyTree[tuple[int]]. So at least internally we make a
            # particular choice of typechecker.
            #
            # Deliberately not using @jaxtyped so that we share the same `memo` as
            # whatever dynamic context we're currently in.
            from jaxtyping._typeguard import typechecked

            @typechecked
            def accepts_leaftype(x: cls.leaftype):  # pyright: ignore[reportInvalidTypeForm]
                pass

            def is_leaftype(x):
                try:
                    accepts_leaftype(x)
                except TypeError:
                    return False
                else:
                    return True

            is_flatten_leaftype = is_check_leaftype = is_leaftype

        set_treeflatten_memo()
        try:
            leaves, structure = tree.tree_flatten(obj, is_leaf=is_flatten_leaftype)
        finally:
            clear_treeflatten_memo()
        if cls.structure is not None:
            if cls.structure.isidentifier():
                try:
                    prev_structure = pytree_memo[cls.structure]
                except KeyError:
                    pytree_memo[cls.structure] = structure
                else:
                    if prev_structure != structure:
                        return False
            else:
                named_pytree = 0
                pieces = cls.structure.split()
                if pieces[0] == "...":
                    pieces = pieces[1:]
                    prefix = False
                    suffix = True
                elif pieces[-1] == "...":
                    pieces = pieces[:-1]
                    prefix = True
                    suffix = False
                else:
                    prefix = False
                    suffix = False
                for identifier in pieces:
                    try:
                        prev_structure = pytree_memo[identifier]
                    except KeyError as e:
                        raise AnnotationError(
                            f"Cannot process composite structure '{cls.structure}' "
                            f"as the structure name {identifier} has not been seen "
                            "before."
                        ) from e
                    # Not using `PyTreeDef.compose` due to JAX bug #18218.
                    prev_pytree = tree.tree_unflatten(
                        prev_structure, [0] * prev_structure.num_leaves
                    )
                    named_pytree = tree.tree_map(lambda _: prev_pytree, named_pytree)
                named_structure = tree.tree_structure(named_pytree)
                if prefix:
                    dummy_pytree = tree.tree_unflatten(structure, [0] * len(leaves))
                    dummy_named = tree.tree_unflatten(
                        named_structure, [0] * named_structure.num_leaves
                    )
                    try:
                        tree.tree_map(lambda _, __: 0, dummy_named, dummy_pytree)
                    except ValueError:
                        return False
                elif suffix:
                    has_structure = lambda x: tree.tree_structure(x) == named_structure
                    dummy_pytree = tree.tree_unflatten(structure, [0] * len(leaves))
                    dummy_leaves = tree.tree_leaves(dummy_pytree, is_leaf=has_structure)
                    if any(not has_structure(x) for x in dummy_leaves):
                        return False
                else:
                    if structure != named_structure:
                        return False

        try:
            for leaf_index, leaf in enumerate(leaves):
                if cls.structure is not None:
                    set_treepath_memo(leaf_index, cls.structure)
                if not is_check_leaftype(leaf):
                    return False
                clear_treepath_memo()
        finally:
            clear_treepath_memo()
        return True

    # Can't return a generic (e.g. _FakePyTree[item]) because generic aliases don't do
    # the custom __instancecheck__ that we want.
    # We can't add that __instancecheck__  via subclassing, e.g.
    # type("PyTree", (Generic[_T],), {}), because dynamic subclassing of typeforms
    # isn't allowed.
    # Likewise we can't do types.new_class("PyTree", (Generic[_T],), {}) because that
    # has __module__ "types", e.g. we get types.PyTree[int].
    @ft.lru_cache(maxsize=None)
    def __getitem__(cls, item):
        if isinstance(item, tuple):
            if len(item) == 2:

                class X(PyTree):  # pyright: ignore[reportRedeclaration]
                    leaftype = item[0]
                    structure = item[1].strip()

                if not isinstance(X.structure, str):
                    raise ValueError(
                        "The structure annotation `struct` in "
                        "`jaxtyping.PyTree[leaftype, struct]` must be be a string, "
                        f"e.g. `jaxtyping.PyTree[leaftype, 'T']`. Got '{X.structure}'."
                    )
                pieces = X.structure.split()
                if len(pieces) == 0:
                    raise ValueError(
                        "The string `struct` in `jaxtyping.PyTree[leaftype, struct]` "
                        "cannot be the empty string."
                    )
                for piece_index, piece in enumerate(pieces):
                    if (piece_index == 0) or (piece_index == len(pieces) - 1):
                        if piece == "...":
                            continue
                    if not piece.isidentifier():
                        raise ValueError(
                            "The string `struct` in "
                            "`jaxtyping.PyTree[leaftype, struct]` must be be a "
                            "whitespace-separated sequence of identifiers, e.g. "
                            "`jaxtyping.PyTree[leaftype, 'T']` or "
                            "`jaxtyping.PyTree[leaftype, 'foo bar']`.\n"
                            "(Here, 'identifier' is used in the same sense as in "
                            "regular Python, i.e. a valid variable name.)\n"
                            f"Got piece '{piece}' in overall structure '{X.structure}'."
                        )

                class Y:
                    pass

                Y.__module__ = "builtins"
                Y.__name__ = repr(X.structure)
                Y.__qualname__ = repr(X.structure)
                name = wl.pformat(_FakePyTree2[X.leaftype, Y], width=9999)
                del Y
            else:
                raise ValueError(
                    "The subscript `foo` in `jaxtyping.PyTree[foo]` must either be a "
                    "leaf type, e.g. `PyTree[int]`, or a 2-tuple of leaf and "
                    "structure, e.g. `PyTree[int, 'T']`. Received a tuple of length "
                    f"{len(item)}."
                )
        else:
            name = wl.pformat(_FakePyTree1[item], width=9999)

            class X(PyTree):
                leaftype = item
                structure = None

        X.__name__ = name
        X.__qualname__ = name
        if getattr(typing, "GENERATING_DOCUMENTATION", "") in {"", "jaxtyping"}:
            X.__module__ = "jaxtyping"
        else:
            X.__module__ = "builtins"
        return X

    def __pdoc__(self, **kwargs):
        if self is PyTree:
            return wl.TextDoc("PyTree")
        else:
            indent = kwargs["indent"]
            docs = [wl.pdoc(self.leaftype, **kwargs)]
            if self.structure is not None:
                docs.append(wl.pdoc(self.structure, **kwargs))
            return wl.bracketed(
                begin=wl.TextDoc("PyTree["),
                docs=docs,
                sep=wl.comma,
                end=wl.TextDoc("]"),
                indent=indent,
            )


# Can't do `class PyTree(Generic[_T]): ...` because we need to override the
# instancecheck for PyTree[foo], but subclassing
# `type(Generic[int])`, i.e. `typing._GenericAlias` is disallowed.
PyTree = _MetaPyTree("PyTree", (), {})
if getattr(typing, "GENERATING_DOCUMENTATION", "") in {"", "jaxtyping"}:
    PyTree.__module__ = "jaxtyping"
else:
    PyTree.__module__ = "builtins"
PyTree.__doc__ = """Represents a PyTree.

Annotations of the following sorts are supported:
```python
a: PyTree
b: PyTree[LeafType]
c: PyTree[LeafType, "T"]
d: PyTree[LeafType, "S T"]
e: PyTree[LeafType, "... T"]
f: PyTree[LeafType, "T ..."]
```

These correspond to:

a. A plain `PyTree` can be used an annotation, in which case `PyTree` is simply a
    suggestively-named alternative to `Any`.
    ([By definition all types are PyTrees.](https://jax.readthedocs.io/en/latest/pytrees.html))

b. `PyTree[LeafType]` denotes a PyTree all of whose leaves match `LeafType`. For
    example, `PyTree[int]` or `PyTree[Union[str, Float32[Array, "b c"]]]`.

c. A structure name can also be passed. In this case
    `jax.tree_util.tree_structure(...)` will be called, and bound to the structure name.
    This can be used to mark that multiple PyTrees all have the same structure:
    ```python
    def f(x: PyTree[int, "T"], y: PyTree[int, "T"]):
        ...
    ```
    Structures are bound to names in the same way as array shape annotations, i.e.
    within the thread-local dynamic context of a [`jaxtyping.jaxtyped`][] decorator.

d. A composite structure can be declared. In this case the variable must have a PyTree
    structure each to the composition of multiple previously-bound PyTree structures.
    For example:
    ```python
    def f(x: PyTree[int, "T"], y: PyTree[int, "S"], z: PyTree[int, "S T"]):
        ...

    x = (1, 2)
    y = {"key": 3}
    z = {"key": (4, 5)}  # structure is the composition of the structures of `y` and `z`
    f(x, y, z)
    ```
    When performing runtime type-checking, all the individual pieces must have already
    been bound to structures, otherwise the composite structure check will throw an error.

e. A structure can begin with a `...`, to denote that the lower levels of the PyTree
    must match the declared structure, but the upper levels can be arbitrary. As in the
    previous case, all named pieces must already have been seen and their structures
    bound.

f. A structure can end with a `...`, to denote that the PyTree must be a prefix of the
    declared structure, but the lower levels can be arbitrary. As in the previous two
    cases, all named pieces must already have been seen and their structures bound.
"""  # noqa: E501



"""Handle re-computation of values of a function given its abstract syntax tree and function frame."""

from __future__ import annotations

import ast
import builtins
import copy
import functools
import inspect
import sys
import uuid
from _ast import If
from collections.abc import Callable, Iterable, Mapping
from typing import Any, TypeVar, cast


class Placeholder:
    """Represent a placeholder for variables local to the lambda such as targets in generator expressions."""

    def __repr__(self) -> str:
        """Represent the placeholder as <Placeholder>."""
        return "<Placeholder>"


PLACEHOLDER = Placeholder()


class FirstExceptionInAll:
    """Represent a first exception case for which an all quantifier does not apply."""

    def __init__(self, result: Any, inputs: tuple[tuple[str, Any]]) -> None:
        """
        Initialize with the given values.

        :param result: value of the evaluation which was not truthy
        :param inputs: all the target loop variables set during the iteration
        """
        self.result = result
        self.inputs = inputs

    def __bool__(self) -> Any:
        """Return the result of the ELT evaluation which invalidated the ``all`` quantifier."""
        return self.result


ContextT = TypeVar("ContextT", bound=ast.expr_context)


class _CollectStoredNamesVisitor(ast.NodeVisitor):
    """Traverse the abstract syntax tree and collect all the names which are stored."""

    def __init__(self) -> None:
        self.names = []  # type: list[str]
        self._name_set = set()  # type: set[str]

    def visit_Name(self, node: ast.Name) -> Any:  # pylint: disable=invalid-name
        """Collect the name if it is in a store context."""
        if isinstance(node.ctx, ast.Store) and node.id not in self._name_set:
            self.names.append(node.id)
            self._name_set.add(node.id)


def _collect_stored_names(nodes: Iterable[ast.expr]) -> list[str]:
    visitor = _CollectStoredNamesVisitor()
    for node in nodes:
        visitor.visit(node)
    return visitor.names


class _CollectNameLoadsVisitor(ast.NodeVisitor):
    """Traverse the abstract syntax tree and collect all the name nodes in Load context."""

    def __init__(self) -> None:
        self.nodes = []  # type: list[ast.expr]

    def visit_Name(self, node: ast.Name) -> Any:  # pylint: disable=invalid-name
        """Collect the name if it is in a load context."""
        if isinstance(node.ctx, ast.Load):
            self.nodes.append(node)


def _collect_name_loads(nodes: Iterable[ast.expr]) -> list[ast.expr]:
    visitor = _CollectNameLoadsVisitor()
    for node in nodes:
        visitor.visit(node)
    return visitor.nodes


# noinspection PyTypeChecker
def _translate_all_expression_to_a_module(
    generator_exp: ast.GeneratorExp,
    generated_function_name: str,
    name_to_value: Mapping[str, Any],
) -> ast.Module:
    """
    Generate the AST of the module to trace an all quantifier on an generator expression.

    :param generator_exp: generator expression to be translated
    :param generated_function_name: UUID of the tracing function to be used in the code
    :param name_to_value:
        mapping of all resolved values to the variable names
        (passed as arguments to the function so that the generation can access them)
    :return: translation to a module
    """
    assert generated_function_name not in name_to_value
    assert not hasattr(builtins, generated_function_name)

    # Collect all the names involved in the generation
    relevant_names = _collect_stored_names(
        generator.target for generator in generator_exp.generators
    )

    assert generated_function_name not in relevant_names

    # Work backwards, from the most-inner block outwards

    result_id = "icontract_tracing_all_result_{}".format(uuid.uuid4().hex)
    result_assignment = ast.Assign(
        targets=[ast.Name(id=result_id, ctx=ast.Store())], value=generator_exp.elt
    )

    exceptional_return = ast.Return(
        ast.Tuple(
            elts=[
                ast.Name(id=result_id, ctx=ast.Load()),
                ast.Tuple(
                    elts=[
                        ast.Tuple(
                            elts=[
                                ast.Constant(value=relevant_name, kind=None),
                                ast.Name(id=relevant_name, ctx=ast.Load()),
                            ],
                            ctx=ast.Load(),
                        )
                        for relevant_name in relevant_names
                    ],
                    ctx=ast.Load(),
                ),
            ],
            ctx=ast.Load(),
        )
    )

    # While happy return shall not be executed, we add it here for robustness in case
    # future refactorings forget to check for that edge case.
    happy_return = ast.Return(
        ast.Tuple(
            elts=[
                ast.Name(id=result_id, ctx=ast.Load()),
                ast.Constant(value=None, kind=None),
            ],
            ctx=ast.Load(),
        )
    )

    critical_if: If = ast.If(
        test=ast.Name(id=result_id, ctx=ast.Load()),
        body=[ast.Pass()],
        orelse=[exceptional_return],
    )

    # Previous inner block to be added as body to the next outer block
    block = None  # type: list[ast.stmt] | None
    for i, comprehension in enumerate(reversed(generator_exp.generators)):
        if i == 0:
            # This is the inner-most comprehension.
            block = [result_assignment, critical_if]
        assert block is not None

        for condition in reversed(comprehension.ifs):
            block = [ast.If(test=condition, body=block, orelse=[])]

        if not comprehension.is_async:
            block = [
                ast.For(
                    target=comprehension.target,
                    iter=comprehension.iter,
                    body=block,
                    orelse=[],
                )
            ]
        else:
            block = [
                ast.AsyncFor(
                    target=comprehension.target,
                    iter=comprehension.iter,
                    body=block,
                    orelse=[],
                )
            ]

    assert block is not None

    block.append(happy_return)

    # Now we are ready to generate the function.

    is_async = any(comprehension.is_async for comprehension in generator_exp.generators)

    args = [ast.arg(arg=name, annotation=None) for name in sorted(name_to_value.keys())]

    if sys.version_info < (3, 5):
        raise NotImplementedError(
            "Python versions below 3.5 not supported, got: {}".format(sys.version_info)
        )

    if not is_async:
        if sys.version_info < (3, 8):
            func_def_node = ast.FunctionDef(
                name=generated_function_name,
                args=ast.arguments(
                    args=args,
                    kwonlyargs=[],
                    kw_defaults=[],
                    defaults=[],
                    vararg=None,
                    kwarg=None,
                ),
                decorator_list=[],
                body=block,
            )  # type: ast.FunctionDef | ast.AsyncFunctionDef

            module_node = ast.Module(body=[func_def_node])
        else:
            func_def_node = ast.FunctionDef(
                name=generated_function_name,
                args=ast.arguments(
                    args=args,
                    posonlyargs=[],
                    kwonlyargs=[],
                    kw_defaults=[],
                    defaults=[],
                    vararg=None,
                    kwarg=None,
                ),
                decorator_list=[],
                body=block,
            )  # type: ignore

            module_node = ast.Module(body=[func_def_node], type_ignores=[])
    else:
        if sys.version_info < (3, 8):
            async_func_def_node = ast.AsyncFunctionDef(
                name=generated_function_name,
                args=ast.arguments(
                    args=args,
                    kwonlyargs=[],
                    kw_defaults=[],
                    defaults=[],
                    vararg=None,
                    kwarg=None,
                ),
                decorator_list=[],
                body=block,
            )

            module_node = ast.Module(body=[async_func_def_node])
        else:
            async_func_def_node = ast.AsyncFunctionDef(
                name=generated_function_name,
                args=ast.arguments(
                    args=args,
                    posonlyargs=[],
                    kwonlyargs=[],
                    kw_defaults=[],
                    defaults=[],
                    vararg=None,
                    kwarg=None,
                ),
                decorator_list=[],
                body=block,
            )  # type: ignore

            module_node = ast.Module(body=[async_func_def_node], type_ignores=[])

    ast.fix_missing_locations(module_node)

    return module_node


# noinspection PyTypeChecker
class Visitor(ast.NodeVisitor):
    """
    Traverse the abstract syntax tree and recompute the values of each node defined by the function frame.

    :ivar recomputed_values: mapping node -> value assigned to each visited node
    :type recomputed_values: Mapping[ast.AST, Any]
    """

    # pylint: disable=invalid-name
    # pylint: disable=missing-docstring

    def __init__(self, variable_lookup: list[Mapping[str, Any]]) -> None:
        """
        Initialize.

        :param variable_lookup: list of lookup tables to look-up the values of the variables, sorted by precedence
        """
        # _name_to_value maps the variable names to variable values.
        # This is important for Load contexts as well as Store contexts in, e.g., named expressions.
        self._name_to_value = dict()  # type: dict[str, Any]

        # Resolve precedence of variable lookups
        for lookup in variable_lookup:
            for name, value in lookup.items():
                if name not in self._name_to_value:
                    self._name_to_value[name] = value

        # value assigned to each visited node
        self.recomputed_values = dict()  # type: dict[ast.AST, Any]

    if sys.version_info < (3, 8):

        def visit_Num(self, node: ast.Num) -> int | float:
            """Recompute the value as the number at the node."""
            result = node.n

            self.recomputed_values[node] = result

            assert isinstance(result, (int, float))
            return result

        def visit_Str(self, node: ast.Str) -> str:
            """Recompute the value as the string at the node."""
            result = node.s

            self.recomputed_values[node] = result
            return result

        def visit_Bytes(self, node: ast.Bytes) -> bytes:
            """Recompute the value as the bytes at the node."""
            result = node.s

            self.recomputed_values[node] = result
            return node.s

        def visit_NameConstant(self, node: ast.NameConstant) -> Any:
            """Forward the node value as a result."""
            self.recomputed_values[node] = node.value
            return node.value

    else:

        def visit_Constant(self, node: ast.Constant) -> Any:
            """Forward the node value as a result."""
            self.recomputed_values[node] = node.value
            return node.value

    if sys.version_info >= (3, 6):

        def visit_FormattedValue(self, node: ast.FormattedValue) -> str | Placeholder:
            """Format the node value."""
            recomputed_format_spec = None  # type: str | Placeholder | None
            if node.format_spec is not None:
                # The following assert serves only documentation purposes so that the code is easier to follow.
                assert isinstance(node.format_spec, ast.JoinedStr)
                recomputed_format_spec = self.visit(node.format_spec)
                assert isinstance(recomputed_format_spec, (str, Placeholder))

            recomputed_value = self.visit(node.value)

            # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
            if recomputed_format_spec is PLACEHOLDER or recomputed_value is PLACEHOLDER:
                return PLACEHOLDER

            fmt = ["{"]
            # See https://docs.python.org/3/library/ast.html#ast.FormattedValue for these
            # constants
            if node.conversion == -1:
                pass
            elif node.conversion == 115:
                fmt.append("!s")
            elif node.conversion == 114:
                fmt.append("!r")
            elif node.conversion == 97:
                fmt.append("!a")
            else:
                raise NotImplementedError(
                    "Unhandled conversion of a formatted value node {!r}: {}".format(
                        node, node.conversion
                    )
                )

            if recomputed_format_spec is not None:
                fmt.append(f":{recomputed_format_spec}")

            fmt.append("}")

            return "".join(fmt).format(recomputed_value)

        def visit_JoinedStr(self, node: ast.JoinedStr) -> str | Placeholder:
            """Visit the values and concatenate them."""
            recomputed_values = [self.visit(value_node) for value_node in node.values]

            # See "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION" above.
            if PLACEHOLDER in recomputed_values:
                return PLACEHOLDER

            joined_str = "".join(recomputed_values)

            self.recomputed_values[node] = joined_str
            return joined_str

    def visit_List(self, node: ast.List) -> list[Any] | Placeholder:
        """Visit the elements and assemble the results into a list."""
        if isinstance(node.ctx, ast.Store):
            raise NotImplementedError("Can not compute the value of a Store on a list")

        recomputed_elts = [self.visit(node=elt) for elt in node.elts]

        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
        if any(recomputed_elt is PLACEHOLDER for recomputed_elt in recomputed_elts):
            return PLACEHOLDER

        self.recomputed_values[node] = recomputed_elts
        return recomputed_elts

    def visit_Tuple(self, node: ast.Tuple) -> tuple[Any, ...] | Placeholder:
        """Visit the elements and assemble the results into a tuple."""
        if isinstance(node.ctx, ast.Store):
            raise NotImplementedError("Can not compute the value of a Store on a tuple")

        recomputed_elts = tuple(self.visit(node=elt) for elt in node.elts)
        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
        if any(recomputed_elt is PLACEHOLDER for recomputed_elt in recomputed_elts):
            return PLACEHOLDER

        self.recomputed_values[node] = recomputed_elts
        return recomputed_elts

    def visit_Set(self, node: ast.Set) -> set[Any] | Placeholder:
        """Visit the elements and assemble the results into a set."""
        recomputed_elts = set(self.visit(node=elt) for elt in node.elts)
        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
        if any(recomputed_elt is PLACEHOLDER for recomputed_elt in recomputed_elts):
            return PLACEHOLDER

        self.recomputed_values[node] = recomputed_elts
        return recomputed_elts

    def visit_Dict(self, node: ast.Dict) -> dict[Any, Any] | Placeholder:
        """Visit keys and values and assemble a dictionary with the results."""
        recomputed_dict = dict()  # type: dict[Any, Any]
        for key, val in zip(node.keys, node.values):
            assert isinstance(key, ast.AST)
            assert isinstance(val, ast.AST)

            recomputed_dict[self.visit(node=key)] = self.visit(node=val)

        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
        if any(
            key is PLACEHOLDER or value is PLACEHOLDER
            for key, value in recomputed_dict.items()
        ):
            return PLACEHOLDER

        self.recomputed_values[node] = recomputed_dict
        return recomputed_dict

    def visit_Name(self, node: ast.Name) -> Any:
        """Load the variable by looking it up in the variable look-up and in the built-ins."""
        if not isinstance(node.ctx, ast.Load):
            raise NotImplementedError(
                "Can only compute a value of Load on a name {}, but got context: {}".format(
                    node.id, node.ctx
                )
            )

        result = None  # type: Any | None

        if node.id in self._name_to_value:
            result = self._name_to_value[node.id]

        if result is None and hasattr(builtins, node.id):
            result = getattr(builtins, node.id)

        if result is None and node.id != "None":
            # The variable refers to a name local of the lambda (e.g., a target in the generator expression).
            # Since we evaluate generator expressions with runtime compilation, None is returned here as a placeholder.
            return PLACEHOLDER

        self.recomputed_values[node] = result
        return result

    def visit_Expr(self, node: ast.Expr) -> Any:
        """Visit the node's ``value``."""
        recomputed_value = self.visit(node=node.value)

        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
        if recomputed_value is PLACEHOLDER:
            return PLACEHOLDER

        self.recomputed_values[node] = recomputed_value
        return recomputed_value

    def visit_UnaryOp(self, node: ast.UnaryOp) -> Any:
        """Visit the node operand and apply the operation on the result."""
        recomputed_operand = self.visit(node=node.operand)

        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
        if recomputed_operand is PLACEHOLDER:
            return PLACEHOLDER

        if isinstance(node.op, ast.UAdd):
            result = +recomputed_operand
        elif isinstance(node.op, ast.USub):
            result = -recomputed_operand
        elif isinstance(node.op, ast.Not):
            result = not recomputed_operand
        elif isinstance(node.op, ast.Invert):
            result = ~recomputed_operand
        else:
            raise NotImplementedError("Unhandled op of {}: {}".format(node, node.op))

        self.recomputed_values[node] = result
        return result

    def visit_BinOp(self, node: ast.BinOp) -> Any:
        """Recursively visit the left and right operand, respectively, and apply the operation on the results."""
        left = self.visit(node=node.left)
        right = self.visit(node=node.right)

        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
        if left is PLACEHOLDER or right is PLACEHOLDER:
            return PLACEHOLDER

        if isinstance(node.op, ast.Add):
            result = left + right
        elif isinstance(node.op, ast.Sub):
            result = left - right
        elif isinstance(node.op, ast.Mult):
            result = left * right
        elif isinstance(node.op, ast.Div):
            result = left / right
        elif isinstance(node.op, ast.FloorDiv):
            result = left // right
        elif isinstance(node.op, ast.Mod):
            result = left % right
        elif isinstance(node.op, ast.Pow):
            result = left**right
        elif isinstance(node.op, ast.LShift):
            result = left << right
        elif isinstance(node.op, ast.RShift):
            result = left >> right
        elif isinstance(node.op, ast.BitOr):
            result = left | right
        elif isinstance(node.op, ast.BitXor):
            result = left ^ right
        elif isinstance(node.op, ast.BitAnd):
            result = left & right
        elif isinstance(node.op, ast.MatMult):
            result = left @ right
        else:
            raise NotImplementedError("Unhandled op of {}: {}".format(node, node.op))

        self.recomputed_values[node] = result
        return result

    def visit_BoolOp(self, node: ast.BoolOp) -> Any:
        """Recursively visit the operands and apply the operation on them."""
        values = [self.visit(value_node) for value_node in node.values]

        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
        if any(value is PLACEHOLDER for value in values):
            return PLACEHOLDER

        if isinstance(node.op, ast.And):
            result = functools.reduce(lambda left, right: left and right, values, True)
        elif isinstance(node.op, ast.Or):
            result = functools.reduce(lambda left, right: left or right, values, True)
        else:
            raise NotImplementedError("Unhandled op of {}: {}".format(node, node.op))

        self.recomputed_values[node] = result
        return result

    def visit_Compare(self, node: ast.Compare) -> Any:
        """Recursively visit the comparators and apply the operations on them."""
        left = self.visit(node=node.left)

        comparators = [self.visit(node=comparator) for comparator in node.comparators]

        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
        if left is PLACEHOLDER or any(
            comparator is PLACEHOLDER for comparator in comparators
        ):
            return PLACEHOLDER

        result = None  # type: Any | None
        for comparator, op in zip(comparators, node.ops):
            if isinstance(op, ast.Eq):
                comparison = left == comparator
            elif isinstance(op, ast.NotEq):
                comparison = left != comparator
            elif isinstance(op, ast.Lt):
                comparison = left < comparator
            elif isinstance(op, ast.LtE):
                comparison = left <= comparator
            elif isinstance(op, ast.Gt):
                comparison = left > comparator
            elif isinstance(op, ast.GtE):
                comparison = left >= comparator
            elif isinstance(op, ast.Is):
                comparison = left is comparator
            elif isinstance(op, ast.IsNot):
                comparison = left is not comparator
            elif isinstance(op, ast.In):
                comparison = left in comparator
            elif isinstance(op, ast.NotIn):
                comparison = left not in comparator
            else:
                raise NotImplementedError("Unhandled op of {}: {}".format(node, op))

            if result is None:
                result = comparison
            else:
                result = result and comparison

            left = comparator

        self.recomputed_values[node] = result
        return result

    def visit_Call(self, node: ast.Call) -> Any:
        """Visit the function and the arguments and finally make the function call with them."""
        func = self.visit(node=node.func)

        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
        if func is PLACEHOLDER:
            return PLACEHOLDER

        if not callable(func):
            raise ValueError(
                "Unexpected call to a non-calllable during the re-computation: {}".format(
                    func
                )
            )

        if inspect.iscoroutinefunction(func):
            raise ValueError(
                (
                    "Unexpected coroutine function {} as a condition of a contract. "
                    "You must specify your own error if the condition of your contract is a coroutine function."
                ).format(func)
            )

        # Short-circuit tracing the all quantifier over a generator expression
        # fmt: off
        if (
                func == builtins.all  # pylint: disable=comparison-with-callable
                and len(node.args) == 1
                and isinstance(node.args[0], ast.GeneratorExp)
        ):
            # fmt: on
            result = self._trace_all_with_generator(func=func, node=node)

            if result is PLACEHOLDER:
                return PLACEHOLDER
        else:
            args = []  # type: list[Any]
            for arg_node in node.args:
                if isinstance(arg_node, ast.Starred):
                    args.extend(self.visit(node=arg_node))
                else:
                    args.append(self.visit(node=arg_node))

            kwargs = dict()  # type: dict[Any, Any]
            for keyword in node.keywords:
                if keyword.arg is None:
                    kw = self.visit(node=keyword.value)
                    for key, val in kw.items():
                        kwargs[key] = val

                else:
                    kwargs[keyword.arg] = self.visit(node=keyword.value)

            # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
            if (any(arg is PLACEHOLDER for arg in args)
                    or any(key is PLACEHOLDER or value is PLACEHOLDER for key, value in kwargs.items())):
                return PLACEHOLDER

            result = func(*args, **kwargs)

        assert result is not PLACEHOLDER

        self.recomputed_values[node] = result
        if inspect.iscoroutine(result):
            raise ValueError(
                ("Unexpected coroutine {} as a result from a call. "
                 "You must specify your own error if the condition of your contract gives a coroutine.").format(result))

        assert node in self.recomputed_values
        return result

    def visit_IfExp(self, node: ast.IfExp) -> Any:
        """Visit the ``test``, and depending on its outcome, the ``body`` or ``orelse``."""
        test = self.visit(node=node.test)

        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
        if test is PLACEHOLDER:
            return PLACEHOLDER

        if test:
            result = self.visit(node=node.body)
        else:
            result = self.visit(node=node.orelse)

        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
        if result is PLACEHOLDER:
            return PLACEHOLDER

        self.recomputed_values[node] = result
        return result

    def visit_Attribute(self, node: ast.Attribute) -> Any:
        """Visit the node's ``value`` and get the attribute from the result."""
        value = self.visit(node=node.value)

        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
        if value is PLACEHOLDER:
            return PLACEHOLDER

        if not isinstance(node.ctx, ast.Load):
            raise NotImplementedError(
                "Can only compute a value of Load on the attribute {}, but got context: {}".format(
                    node.attr, node.ctx
                )
            )

        result = getattr(value, node.attr)

        self.recomputed_values[node] = result
        return result

    if sys.version_info >= (3, 8):

        def visit_NamedExpr(self, node: ast.NamedExpr) -> Any:
            """Visit the node's ``value`` and assign it to both this node and the target."""
            value = self.visit(node=node.value)

            # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
            if value is PLACEHOLDER:
                return PLACEHOLDER

            self.recomputed_values[node] = value

            target = node.target

            if not isinstance(target.ctx, ast.Store):
                raise NotImplementedError(
                    "Expected Store context in the target of a named expression, but got: {}".format(
                        target.ctx
                    )
                )

            self._name_to_value[target.id] = value

            return value

    if sys.version_info < (3, 9):

        def visit_Index(self, node: ast.Index) -> Any:
            """Visit the node's ``value``."""
            result = self.visit(node=node.value)

            # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
            if result is PLACEHOLDER:
                return PLACEHOLDER

            self.recomputed_values[node] = result
            return result

    def visit_Slice(self, node: ast.Slice) -> slice | Placeholder:
        """Visit ``lower``, ``upper`` and ``step`` and recompute the node as a ``slice``."""
        lower = None  # type: int | Placeholder | None
        if node.lower is not None:
            lower = self.visit(node=node.lower)

        upper = None  # type: int | Placeholder | None
        if node.upper is not None:
            upper = self.visit(node=node.upper)

        step = None  # type: int | Placeholder | None
        if node.step is not None:
            step = self.visit(node=node.step)

        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
        if lower is PLACEHOLDER or upper is PLACEHOLDER or step is PLACEHOLDER:
            return PLACEHOLDER

        result = slice(lower, upper, step)

        self.recomputed_values[node] = result
        return result

    if sys.version_info < (3, 9):

        def visit_ExtSlice(self, node: ast.ExtSlice) -> tuple[Any, ...] | Placeholder:
            """Visit each dimension of the advanced slicing and assemble the dimensions in a tuple."""
            result = tuple(self.visit(node=dim) for dim in node.dims)

            if any(value is PLACEHOLDER for value in result):
                return PLACEHOLDER

            self.recomputed_values[node] = result
            return result

    def visit_Subscript(self, node: ast.Subscript) -> Any:
        """Visit the ``slice`` and a ``value`` and get the element."""
        value = self.visit(node=node.value)
        a_slice = self.visit(node=node.slice)

        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
        if value is PLACEHOLDER or a_slice is PLACEHOLDER:
            return PLACEHOLDER

        result = value[a_slice]
        self.recomputed_values[node] = result
        return result

    def _trace_all_with_generator(
        self, func: Callable[..., Any], node: ast.Call
    ) -> Any:
        """Re-write the all call with for loops to trace the first offending item, if any."""
        assert func == builtins.all  # pylint: disable=comparison-with-callable
        assert len(node.args) == 1
        assert isinstance(node.args[0], ast.GeneratorExp)

        # Try the happy path first

        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION"
        recomputed_arg = self.visit(node=node.args[0])
        if recomputed_arg is PLACEHOLDER:
            return PLACEHOLDER

        result = func(*(self.visit(node=node.args[0]),))
        if result:
            return result

        # The all quantifier has not been satisfied. We need to re-trace it.
        # To that end, we translate the generator expression to a tracing function and
        # execute it.

        generator_exp = node.args[0]
        assert isinstance(generator_exp, ast.GeneratorExp)

        generated_function_name = "icontract_tracing_all_with_generator_expr_{}".format(
            uuid.uuid4().hex
        )

        module_node = _translate_all_expression_to_a_module(
            generator_exp=generator_exp,
            generated_function_name=generated_function_name,
            name_to_value=self._name_to_value,
        )

        # In case you want to debug the generated function at this point,
        # you probably want to use ``astor`` module to generate the source code
        # based on the ``module_node``.

        code = compile(source=module_node, filename="<ast>", mode="exec")

        module_locals = {}  # type: dict[str, Any]
        module_globals = {}  # type: dict[str, Any]
        exec(code, module_globals, module_locals)  # pylint: disable=exec-used

        generated_func = module_locals[generated_function_name]

        result, inputs = generated_func(**self._name_to_value)

        assert not bool(result), "Expected the unhappy path here"
        assert isinstance(inputs, tuple)
        assert all(
            isinstance(item, tuple) and len(item) == 2 and isinstance(item[0], str)
            for item in inputs
        )

        return FirstExceptionInAll(
            result=result, inputs=cast(tuple[tuple[str, Any]], inputs)
        )

    def _execute_comprehension(
        self,
        node: ast.ListComp | ast.SetComp | ast.GeneratorExp | ast.DictComp,
    ) -> Any:
        """Compile the generator or comprehension from the node and execute the compiled code."""
        # Please see "NOTE ABOUT NAME 🠒 VALUE STACKING".
        if any(value is PLACEHOLDER for value in self._name_to_value.values()):
            return PLACEHOLDER

        args = [
            ast.arg(arg=name, annotation=None)
            for name in sorted(self._name_to_value.keys())
        ]

        if sys.version_info < (3,):
            raise NotImplementedError(
                "Python versions below 3 not supported, got: {}".format(
                    sys.version_info
                )
            )

        if sys.version_info < (3, 8):
            func_def_node = ast.FunctionDef(
                name="generator_expr",
                args=ast.arguments(
                    args=args, kwonlyargs=[], kw_defaults=[], defaults=[]
                ),
                decorator_list=[],
                body=[ast.Return(node)],
            )

            module_node = ast.Module(body=[func_def_node])
        else:
            func_def_node = ast.FunctionDef(
                name="generator_expr",
                args=ast.arguments(
                    args=args,
                    posonlyargs=[],
                    kwonlyargs=[],
                    kw_defaults=[],
                    defaults=[],
                ),
                body=[ast.Return(node)],
                decorator_list=[],
            )  # type: ignore

            module_node = ast.Module(body=[func_def_node], type_ignores=[])

        ast.fix_missing_locations(module_node)

        code = compile(source=module_node, filename="<ast>", mode="exec")

        module_locals = {}  # type: dict[str, Any]
        module_globals = {}  # type: dict[str, Any]
        exec(code, module_globals, module_locals)  # pylint: disable=exec-used

        generator_expr_func = module_locals["generator_expr"]

        return generator_expr_func(**self._name_to_value)

    def visit_GeneratorExp(self, node: ast.GeneratorExp) -> Any:
        """Compile the generator expression as a function and call it."""
        # NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION:
        # Re-computing the comprehensions would be too slow. Therefore we re-compile the comprehension and call
        # the compiled code directly.
        #
        # However, we still want to report the values of the variables unrelated to the comprehension back
        # to the user. Therefore we introduce PLACEHOLDER's and propagate them through re-computation in all
        # the visit methods.

        # The following visits propagate the visitation to descendant nodes.
        # However, as we re-compute the comprehension *via* re-compilation & execution,
        # the results of the visits are all PLACEHOLDER's.

        # NOTE ABOUT NAME 🠒 VALUE STACKING:
        # We need to make a copy of name 🠒 value mapping since we need to add targets as placeholders
        # while we visit the comprehension. For example, as we compute comprehensions through re-compilation
        # and not through manual re-computation, we can not re-compute nested comprehensions.
        #
        # However, when our visit of comprehension is finished, the targets are not valid any more,
        # so we need to remove them from the mapping.
        #
        # Finally, we compute the comprehension with the original name 🠒 value mapping by using
        # re-compilation. This final step is skipped if any of the names involved in the comprehension are
        # PLACEHOLDER's.

        old_name_to_value = copy.copy(self._name_to_value)
        for target_name in _collect_stored_names(
            [generator.target for generator in node.generators]
        ):
            self._name_to_value[target_name] = PLACEHOLDER

        self.visit(node.elt)

        for generator in node.generators:
            self.visit(generator.iter)

            for generator_if in generator.ifs:
                self.visit(generator_if)

        self._name_to_value = old_name_to_value

        result = self._execute_comprehension(node=node)

        # Do not set the computed value of the node since its representation would be non-informative.
        return result

    def visit_ListComp(self, node: ast.ListComp) -> Any:
        """Compile the list comprehension as a function and call it."""
        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION".
        # The note also explains why we do not use the result of the following visits.

        # Please see "NOTE ABOUT NAME 🠒 VALUE STACKING".
        old_name_to_value = copy.copy(self._name_to_value)
        for target_name in _collect_stored_names(
            [generator.target for generator in node.generators]
        ):
            self._name_to_value[target_name] = PLACEHOLDER

        self.visit(node.elt)

        for generator in node.generators:
            self.visit(generator.iter)

            for generator_if in generator.ifs:
                self.visit(generator_if)

        self._name_to_value = old_name_to_value

        result = self._execute_comprehension(node=node)

        if result is not PLACEHOLDER:
            self.recomputed_values[node] = result

        return result

    def visit_SetComp(self, node: ast.SetComp) -> Any:
        """Compile the set comprehension as a function and call it."""
        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION".
        # The note also explains why we do not use the result of the following visits.

        # Please see "NOTE ABOUT NAME 🠒 VALUE STACKING".
        old_name_to_value = copy.copy(self._name_to_value)
        for target_name in _collect_stored_names(
            [generator.target for generator in node.generators]
        ):
            self._name_to_value[target_name] = PLACEHOLDER

        self.visit(node.elt)

        for generator in node.generators:
            self.visit(generator.iter)

            for generator_if in generator.ifs:
                self.visit(generator_if)

        self._name_to_value = old_name_to_value

        result = self._execute_comprehension(node=node)

        if result is not PLACEHOLDER:
            self.recomputed_values[node] = result

        return result

    def visit_DictComp(self, node: ast.DictComp) -> Any:
        """Compile the dictionary comprehension as a function and call it."""
        # Please see "NOTE ABOUT PLACEHOLDERS AND RE-COMPUTATION".
        # The note also explains why we do not use the result of the following visits.

        # Please see "NOTE ABOUT NAME 🠒 VALUE STACKING".
        old_name_to_value = copy.copy(self._name_to_value)
        for target_name in _collect_stored_names(
            [generator.target for generator in node.generators]
        ):
            self._name_to_value[target_name] = PLACEHOLDER

        self.visit(node.key)
        self.visit(node.value)

        for generator in node.generators:
            self.visit(generator.iter)

            for generator_if in generator.ifs:
                self.visit(generator_if)

        self._name_to_value = old_name_to_value

        result = self._execute_comprehension(node=node)

        if result is not PLACEHOLDER:
            self.recomputed_values[node] = result

        return result

    def visit_Lambda(self, node: ast.Lambda) -> Callable[..., Any]:
        """Do not support inline lambda until there is a feature request since this is quite tricky to implement."""
        raise NotImplementedError(
            "Re-computation of in-line lambda functions is not supported since it is quite tricky to implement and "
            "we decided to implement it only once there is a real need for it. "
            "Please make a feature request on https://github.com/Parquery/icontract"
        )

    def visit_Return(self, node: ast.Return) -> Any:
        """Raise an exception that this node is unexpected."""
        raise AssertionError(
            "Unexpected return node during the re-computation: {}".format(
                ast.dump(node)
            )
        )

    def generic_visit(self, node: ast.AST) -> None:
        """Raise an exception that this node has not been handled."""
        raise NotImplementedError(
            "Unhandled re-computation of the node: {} {}".format(type(node), node)
        )



# Convenience functions for common patterns
from __future__ import annotations

from collections.abc import Callable

from ._config import ROOT_PATH, enabled


def assert_(
    condition: bool | Callable[[], bool],
    message: str,
    *,
    path: str = ROOT_PATH,
) -> None:
    """Hierarchical debug assertion.

    Args:
        condition: Condition to check
        message: Error message if assertion fails
        path: Debug path for hierarchical control.

    Examples:
        assert_(tensor.shape[0] > 0, "Batch size must be positive", path="training.data")
    """
    if not enabled(path):
        return

    if callable(condition):
        condition = condition()

    if not condition:
        raise AssertionError(f"Assertion failed at {path}: {message}")



from __future__ import annotations

import contextlib
import fnmatch
import os
import tempfile
import weakref
from collections.abc import Callable, Mapping
from dataclasses import dataclass
from functools import wraps
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generic, Literal, Union, cast, overload

import numpy as np
from typing_extensions import Never, ParamSpec, TypeAliasType, TypeVar, override
from uuid_extensions import uuid7str

from ..collections import apply_to_collection

if not TYPE_CHECKING:
    try:
        import torch  # pyright: ignore[reportMissingImports]

        Tensor = torch.Tensor
        _torch_installed = True
    except ImportError:
        torch = None
        _torch_installed = False

        Tensor = Never
else:
    import torch  # pyright: ignore[reportMissingImports]

    Tensor = TypeAliasType("Tensor", torch.Tensor)
    _torch_installed: Literal[True] = True

log = getLogger(__name__)

# Updated to include Any for arbitrary types
Value = TypeAliasType(
    "Value", Union[int, float, complex, bool, str, np.ndarray, Tensor, Any, None]
)
ValueOrLambda = TypeAliasType("ValueOrLambda", Union[Value, Callable[..., Value]])


def _torch_is_scripting() -> bool:
    if not _torch_installed:
        return False

    return torch.jit.is_scripting()


def _to_numpy(activation: Value) -> np.ndarray:
    # Make sure it's not `None`
    if activation is None:
        raise ValueError("Activation should not be `None`")

    if isinstance(activation, (int, float, complex, str, bool)):
        return np.array(activation)
    elif isinstance(activation, np.ndarray):
        return activation
    elif _torch_installed and isinstance(activation, torch.Tensor):
        activation_ = activation.detach()
        if activation_.is_floating_point():
            # NOTE: We need to convert to float32 because [b]float16 is not supported by numpy
            activation_ = activation_.float()
        return activation_.cpu().numpy()
    else:
        # Handle arbitrary objects using numpy object dtype
        return np.array(activation, dtype=object)


T = TypeVar("T", infer_variance=True)


# A wrapper around weakref.ref that allows for primitive types
# To get around errors like:
# TypeError: cannot create weak reference to 'int' object
class WeakRef(Generic[T]):
    _ref: Callable[[], T] | None

    def __init__(self, obj: T):
        try:
            self._ref = cast(Callable[[], T], weakref.ref(obj))
        except TypeError as e:
            if "cannot create weak reference" not in str(e):
                raise
            self._ref = lambda: obj

    def __call__(self) -> T:
        if self._ref is None:
            raise RuntimeError("WeakRef is deleted")
        return self._ref()

    def delete(self):
        del self._ref
        self._ref = None


@dataclass
class Activation:
    name: str
    ref: WeakRef[ValueOrLambda] | None
    transformed: np.ndarray | None = None

    def __post_init__(self):
        # Update the `name` to replace `/` with `.`
        self.name = self.name.replace("/", ".")

    def __call__(self) -> np.ndarray | None:
        # If we have a transformed value, we return it
        if self.transformed is not None:
            return self.transformed

        if self.ref is None:
            raise RuntimeError("Activation is deleted")

        # If we have a lambda, we need to call it
        unrwapped_ref = self.ref()
        activation = unrwapped_ref
        if callable(unrwapped_ref):
            activation = unrwapped_ref()

        # If we have a `None`, we return early
        if activation is None:
            return None

        if _torch_installed:
            activation = apply_to_collection(activation, Tensor, _to_numpy)
        activation = _to_numpy(activation)

        # Set the transformed value
        self.transformed = activation

        # Delete the reference
        self.ref.delete()
        del self.ref
        self.ref = None

        return self.transformed

    @classmethod
    def from_value_or_lambda(cls, name: str, value_or_lambda: ValueOrLambda):
        return cls(name, WeakRef(value_or_lambda))

    @classmethod
    def from_dict(cls, d: Mapping[str, ValueOrLambda]):
        return [cls.from_value_or_lambda(k, v) for k, v in d.items()]


Transform = Callable[[Activation], Mapping[str, ValueOrLambda]]


def _ensure_supported():
    try:
        import torch.distributed as dist  # type: ignore

        if dist.is_initialized() and dist.get_world_size() > 1:
            raise RuntimeError("Only single GPU is supported at the moment")
    except ImportError:
        pass


P = ParamSpec("P")


def _ignore_if_scripting(fn: Callable[P, None]) -> Callable[P, None]:
    @wraps(fn)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> None:
        if _torch_is_scripting():
            return

        _ensure_supported()
        fn(*args, **kwargs)

    return wrapper


class _Saver:
    def __init__(
        self,
        save_dir: Path,
        prefixes_fn: Callable[[], list[str]],
        *,
        filters: list[str] | None = None,
    ):
        # Create a directory under `save_dir` by autoincrementing
        # (i.e., every activation save context, we create a new directory)
        # The id = the number of activation subdirectories
        self._id = sum(1 for subdir in save_dir.glob("*") if subdir.is_dir())
        save_dir.mkdir(parents=True, exist_ok=True)

        # Add a .activationbase file to the save_dir to indicate that this is an activation base
        (save_dir / ".activationbase").touch(exist_ok=True)

        self._save_dir = save_dir / f"{self._id:04d}"
        # Make sure `self._save_dir` does not exist and create it
        self._save_dir.mkdir(exist_ok=False)

        self._prefixes_fn = prefixes_fn
        self._filters = filters

    def _save_activation(self, activation: Activation):
        # If the activation value is `None`, we skip it.
        if (activation_value := activation()) is None:
            return

        # Save the activation to self._save_dir / name / {id}.npz, where id is an auto-incrementing integer
        file_name = ".".join(self._prefixes_fn() + [activation.name])
        path = self._save_dir / file_name
        path.mkdir(exist_ok=True, parents=True)

        # Get the next id and save the activation
        id = len(list(path.glob("*.npy")))
        np.save(path / f"{id:04d}.npy", activation_value)

    @_ignore_if_scripting
    def save(
        self,
        acts: dict[str, ValueOrLambda] | None = None,
        /,
        **kwargs: ValueOrLambda,
    ):
        kwargs.update(acts or {})

        # Build activations
        activations = Activation.from_dict(kwargs)

        for activation in activations:
            # Make sure name matches at least one filter if filters are specified
            if self._filters is not None and all(
                not fnmatch.fnmatch(activation.name, f) for f in self._filters
            ):
                continue

            # Save the current activation
            self._save_activation(activation)

        del activations


class ActSaveProvider:
    _saver: _Saver | None = None
    _prefixes: list[str] = []
    _disable_count: int = 0

    @property
    def is_initialized(self) -> bool:
        """Returns True if ActSave.enable() has been called and not subsequently disabled."""
        return self._saver is not None

    @property
    def is_enabled(self) -> bool:
        """Returns True if ActSave is currently active and will save activations."""
        return self.is_initialized and self._disable_count == 0

    def enable(self, save_dir: Path | None = None):
        """
        Initializes the saver with the given configuration and save directory.

        Args:
            save_dir (Path): The directory where the saved files will be stored.
        """
        if self._saver is not None:
            log.warning("ActSave is already enabled")
            return

        if save_dir is None:
            save_dir = Path(tempfile.gettempdir()) / f"actsave-{uuid7str()}"
            log.warning(
                f"ActSave: Using temporary directory {save_dir} for activations."
            )
        else:
            log.info(f"ActSave enabled. Saving to {save_dir}")
        self._saver = _Saver(save_dir, lambda: self._prefixes)

    def disable(self):
        """
        Disables the actsaver.
        """
        if self._saver is None:
            log.warning("ActSave is already disabled")
            return

        del self._saver
        self._saver = None

    @contextlib.contextmanager
    def enabled(self, save_dir: Path | None = None):
        """
        Context manager that enables the actsave functionality with the specified configuration.

        Args:
            save_dir (Path): The directory where the saved files will be stored.
        """
        if self._saver is not None:
            log.warning("ActSave is already enabled")
            yield
            return

        self.enable(save_dir)
        try:
            yield
        finally:
            self.disable()

    @override
    def __init__(self):
        super().__init__()

        self._saver = None
        self._prefixes = []
        self._disable_count = 0

        # Check for environment variable `ACTSAVE` to automatically enable saving.
        # If set to "1" or "true" (case-insensitive), activations are saved to a temporary directory.
        # If set to a path, activations are saved to that path.
        if env_var := os.environ.get("ACTSAVE"):
            log.info(
                f"`ACTSAVE={env_var}` detected, attempting to auto-enable activation saving."
            )
            if env_var.lower() in ("1", "true"):
                self.enable()
            else:
                self.enable(Path(env_var))

    @contextlib.contextmanager
    def disabled(self, condition: bool | Callable[[], bool] = True):
        """
        Context manager to temporarily disable activation saving.

        Args:
            condition (bool | Callable[[], bool], optional):
                If True or a callable returning True, saving is disabled within this context.
                Defaults to True.
        """
        if _torch_is_scripting():
            yield
            return

        should_disable = condition() if callable(condition) else condition
        if should_disable:
            self._disable_count += 1

        try:
            yield
        finally:
            if should_disable:
                self._disable_count -= 1
                if self._disable_count < 0:  # Should not happen
                    log.warning("ActSave disable count went below zero.")
                    self._disable_count = 0

    @contextlib.contextmanager
    def context(self, label: str):
        """
        A context manager that adds a label to the current context.

        Args:
            label (str): The label for the context.
        """
        if _torch_is_scripting():
            yield
            return

        if not self.is_enabled:
            yield
            return

        _ensure_supported()

        log.debug(f"Entering ActSave context {label}")
        self._prefixes.append(label)
        try:
            yield
        finally:
            _ = self._prefixes.pop()

    prefix = context

    @overload
    def __call__(
        self,
        acts: dict[str, ValueOrLambda] | None = None,
        /,
        **kwargs: ValueOrLambda,
    ):
        """
        Saves the activations to disk.

        Args:
            acts (dict[str, ValueOrLambda] | None, optional): A dictionary of acts. Defaults to None.
            **kwargs (ValueOrLambda): Additional keyword arguments.

        Returns:
            None

        """
        ...

    @overload
    def __call__(self, acts: Callable[[], dict[str, ValueOrLambda]], /):
        """
        Saves the activations to disk.

        Args:
            acts (Callable[[], dict[str, ValueOrLambda]]): A callable that returns a dictionary of acts.
            **kwargs (ValueOrLambda): Additional keyword arguments.

        Returns:
            None

        """
        ...

    def __call__(
        self,
        acts: (
            dict[str, ValueOrLambda] | Callable[[], dict[str, ValueOrLambda]] | None
        ) = None,
        /,
        **kwargs: ValueOrLambda,
    ):
        if _torch_is_scripting():
            return

        if not self.is_enabled:
            return

        # Ensure _saver is not None, which is guaranteed by is_enabled but mypy needs help
        assert self._saver is not None

        if acts is not None and callable(acts):
            acts = acts()
        self._saver.save(acts, **kwargs)

    save = __call__


ActSave = ActSaveProvider()



# Copyright The PyTorch Lightning team.
# Licensed under the Apache License, Version 2.0 (the "License");
#     http://www.apache.org/licenses/LICENSE-2.0
#
from __future__ import annotations

import dataclasses
from collections import OrderedDict, defaultdict
from collections.abc import Callable, Mapping, Sequence
from copy import deepcopy
from typing import Any, cast


def is_namedtuple(obj: object) -> bool:
    """Check if object is type nametuple."""
    # https://github.com/pytorch/pytorch/blob/v1.8.1/torch/nn/parallel/scatter_gather.py#L4-L8
    return (
        isinstance(obj, tuple) and hasattr(obj, "_asdict") and hasattr(obj, "_fields")
    )


def is_dataclass_instance(obj: object) -> bool:
    """Check if object is dataclass."""
    # https://docs.python.org/3/library/dataclasses.html#module-level-decorators-classes-and-functions
    return dataclasses.is_dataclass(obj) and not isinstance(obj, type)


def apply_to_collection(
    data: Any,
    dtype: type | Any | tuple[type | Any],
    function: Callable,
    *args: Any,
    wrong_dtype: type | tuple[type, ...] | None = None,
    include_none: bool = True,
    allow_frozen: bool = False,
    **kwargs: Any,
) -> Any:
    """Recursively applies a function to all elements of a certain dtype.

    Args:
        data: the collection to apply the function to
        dtype: the given function will be applied to all elements of this dtype
        function: the function to apply
        *args: positional arguments (will be forwarded to calls of ``function``)
        wrong_dtype: the given function won't be applied if this type is specified and the given collections
            is of the ``wrong_dtype`` even if it is of type ``dtype``
        include_none: Whether to include an element if the output of ``function`` is ``None``.
        allow_frozen: Whether not to error upon encountering a frozen dataclass instance.
        **kwargs: keyword arguments (will be forwarded to calls of ``function``)

    Returns:
        The resulting collection
    """
    # Breaking condition
    if isinstance(data, dtype) and (
        wrong_dtype is None or not isinstance(data, wrong_dtype)
    ):
        return function(data, *args, **kwargs)

    elem_type = type(data)

    # Recursively apply to collection items
    if isinstance(data, Mapping):
        out = []
        for k, v in data.items():
            v = apply_to_collection(
                v,
                dtype,
                function,
                *args,
                wrong_dtype=wrong_dtype,
                include_none=include_none,
                allow_frozen=allow_frozen,
                **kwargs,
            )
            if include_none or v is not None:
                out.append((k, v))
        if isinstance(data, defaultdict):
            return elem_type(data.default_factory, OrderedDict(out))
        return elem_type(OrderedDict(out))

    is_namedtuple_ = is_namedtuple(data)
    is_sequence = isinstance(data, Sequence) and not isinstance(data, str)
    if is_namedtuple_ or is_sequence:
        out = []
        for d in data:
            v = apply_to_collection(
                d,
                dtype,
                function,
                *args,
                wrong_dtype=wrong_dtype,
                include_none=include_none,
                allow_frozen=allow_frozen,
                **kwargs,
            )
            if include_none or v is not None:
                out.append(v)
        return elem_type(*out) if is_namedtuple_ else elem_type(out)

    if is_dataclass_instance(data):
        data = cast(Any, data)
        # make a deepcopy of the data,
        # but do not deepcopy mapped fields since the computation would
        # be wasted on values that likely get immediately overwritten
        fields = {}
        memo = {}
        for field in dataclasses.fields(data):
            field_value = getattr(data, field.name)
            fields[field.name] = (field_value, field.init)
            memo[id(field_value)] = field_value
        result = deepcopy(data, memo=memo)
        # apply function to each field
        for field_name, (field_value, field_init) in fields.items():
            v = None
            if field_init:
                v = apply_to_collection(
                    field_value,
                    dtype,
                    function,
                    *args,
                    wrong_dtype=wrong_dtype,
                    include_none=include_none,
                    allow_frozen=allow_frozen,
                    **kwargs,
                )
            if not field_init or (not include_none and v is None):  # retain old value
                v = getattr(data, field_name)
            try:
                setattr(result, field_name, v)
            except dataclasses.FrozenInstanceError as e:
                if allow_frozen:
                    # Quit early if we encounter a frozen data class; return `result` as is.
                    break
                raise ValueError(
                    "A frozen dataclass was passed to `apply_to_collection` but this is not allowed."
                ) from e
        return result

    # data is neither of dtype, nor a collection
    return data



from __future__ import annotations

import os
from collections.abc import Mapping, MutableMapping
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Final, TypeAlias, TypedDict

from typing_extensions import Self, assert_never
from typing_extensions import override as override_


class DebugConfig(TypedDict, total=False):
    enabled: bool


_SENTINEL: Final = object()
ROOT_PATH: Final = ""

ValueType: TypeAlias = DebugConfig | bool | None


def _config_for_value(value: ValueType) -> DebugConfig | object:
    """Convert a bool | None value to either a DebugConfig or _SENTINEL."""
    match value:
        case None:
            return _SENTINEL
        case bool():
            return DebugConfig(enabled=value)
        case Mapping():
            return value
        case _:
            assert_never(value)


def _default_enabled():
    return __debug__


class _DebugNode:
    __slots__ = ("_path", "_parent", "_var")
    _registry: MutableMapping[str, Self] = {}

    @classmethod
    def node(cls, path: str) -> Self:
        """Create or fetch the node for path."""
        if path in cls._registry:
            return cls._registry[path]

        if not path:  # root
            parent = None
        else:
            parent_path, _, _ = path.rpartition(".")
            parent = cls.node(parent_path)

        self = super().__new__(cls)
        self._init(path, parent)
        cls._registry[path] = self
        return self

    def _init(self, path: str, parent: "_DebugNode | None") -> None:
        self._path = path
        self._parent = parent
        # A ContextVar holding either DebugConfig or _SENTINEL for inheritance
        default: DebugConfig | object
        if parent is None:
            # Only the root uses __debug__ as default
            default = DebugConfig(enabled=_default_enabled())
        else:
            default = _SENTINEL

        self._var: ContextVar[DebugConfig | object] = ContextVar(
            f"debug:{path or '<root>'}", default=default
        )

    def _config(self) -> DebugConfig:
        val = self._var.get()
        if val is _SENTINEL:
            return (
                self._parent._config() if self._parent else DebugConfig(enabled=False)
            )
        return val  # type: ignore[return-value]

    @property
    def config(self) -> DebugConfig:
        """Return the effective configuration for this node."""
        return self._config()

    def __bool__(self) -> bool:
        """Return effective enabled flag."""
        return self.config.get("enabled", False)

    def enabled(self) -> bool:
        """Return effective enabled flag."""
        return bool(self)

    def set(self, value: ValueType) -> None:
        """Set value. None means inherit parent again."""
        self._var.set(_config_for_value(value))

    @contextmanager
    def override(self, value: ValueType):
        """Temporarily override value within current context."""
        token = self._var.set(_config_for_value(value))
        try:
            yield
        finally:
            self._var.reset(token)

    @override_
    def __repr__(self) -> str:
        return f"<DebugNode {self._path!r} enabled={bool(self)}>"


# Convenience layer
def node(path: str = ROOT_PATH) -> _DebugNode:
    """Return the (singleton) node object for path."""
    return _DebugNode.node(path)


def config(path: str = ROOT_PATH):
    """Return the effective configuration for path."""
    return node(path).config


def enabled(path: str = ROOT_PATH) -> bool:
    """Cheap functional check: if enabled("foo.bar"):"""
    return bool(node(path))


@contextmanager
def override(value: ValueType, path: str = ROOT_PATH):
    """Temporarily override path within the current async context."""
    with node(path).override(value):
        yield


# Optional environment bootstrap
def _bootstrap_from_env(prefix: str = "NSHUTILS_DEBUG_") -> None:
    """
    Support NSHUTILS_DEBUG= (root) or NSHUTILS_DEBUG_TRAINING_MASKS= style flags.

    "1", "true", "yes", "on" -> True
    "0", "false", "no", "off" -> False
    """
    truthy = {"1", "true", "yes", "on"}
    falsy = {"0", "false", "no", "off"}
    for key, raw in os.environ.items():
        if not key.startswith(prefix):
            continue
        path = key[len(prefix) :].lower().replace("_", ".")  # env uses '_' separator
        if raw.lower() in truthy:
            node(path).set(True)
        elif raw.lower() in falsy:
            node(path).set(False)


_bootstrap_from_env()



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING

from ._config import ROOT_PATH, override

try:
    import lightning.pytorch as pl  # pyright: ignore[reportMissingImports]

    class SanityDebugFlagCallback(pl.Callback):
        """Callback to enable debug contexts during sanity check."""

        def __init__(
            self,
            paths: list[str] | None = None,
        ):
            self._paths = paths or [ROOT_PATH, "sanity_check"]
            self._contexts: list[contextlib.AbstractContextManager] = []

        def on_sanity_check_start(
            self, trainer: pl.Trainer, pl_module: pl.LightningModule
        ):
            """Enable debug contexts during sanity check."""
            for path in self._paths:
                context = override({"enabled": True}, path=path)
                self._contexts.append(context)

        def on_sanity_check_end(
            self, trainer: pl.Trainer, pl_module: pl.LightningModule
        ):
            """Restore debug contexts after sanity check."""
            for context in reversed(self._contexts):
                context.__exit__(None, None, None)
            self._contexts.clear()

except ImportError:
    if not TYPE_CHECKING:

        class SanityDebugFlagCallback:
            def __init__(self, *args, **kwargs):
                raise ImportError(
                    "PyTorch Lightning is not installed. Please install it to use SanityDebugFlagCallback."
                )



"""Handle representations necessary for informative error messages."""

from __future__ import annotations

import ast
import inspect
import re
import reprlib
import sys
import textwrap
import uuid
from collections.abc import Callable, Mapping, MutableMapping
from typing import Any, cast

import asttokens.asttokens

from . import _recompute
from ._globals import AnyCallable
from ._types import Contract


def _representable(value: Any) -> bool:
    """
    Check whether we want to represent the value in the error message on contract breach.

    We do not want to represent classes, methods, modules and functions.

    :param value: value related to an AST node
    :return: True if we want to represent it in the violation error
    """
    return (
        not inspect.isclass(value)
        and not inspect.isfunction(value)
        and not inspect.ismethod(value)
        and not inspect.ismodule(value)
        and not inspect.isbuiltin(value)
    )


class Visitor(ast.NodeVisitor):
    """Traverse the abstract syntax tree and collect the representations of the selected nodes."""

    # pylint: disable=invalid-name
    # pylint: disable=missing-docstring

    def __init__(
        self,
        recomputed_values: Mapping[ast.AST, Any],
        variable_lookup: list[Mapping[str, Any]],
        atok: asttokens.asttokens.ASTTokens,
    ) -> None:
        """
        Initialize.

        :param recomputed_values: AST node of a condition function -> value associated with the node
        :param variable_lookup:
            list of lookup tables to look-up the values of the variables, sorted by precedence.
            The visitor needs it here to check whether we overrode a built-in variable (like ``id``).
        :param atok: parsed AST tree and tokens with additional positions in source code

        """
        self._recomputed_values = recomputed_values
        self._variable_lookup = variable_lookup
        self.reprs = dict()  # type: MutableMapping[str, str]
        self._atok = atok

    if sys.version_info >= (3, 6):

        def visit_JoinedStr(self, node: ast.JoinedStr) -> None:
            """Show the whole joined strings without descending into the values."""
            if node in self._recomputed_values:
                value = self._recomputed_values[node]

                if _representable(value=value):
                    text = self._atok.get_text(node)
                    self.reprs[text] = value

    def visit_Name(self, node: ast.Name) -> None:
        """
        Resolve the name from the variable look-up and the built-ins.

        Due to possible branching (e.g., If-expressions), some nodes might lack the recomputed values. These nodes
        are ignored.
        """
        if node in self._recomputed_values:
            value = self._recomputed_values[node]

            # Check if it is a non-built-in
            is_builtin = True
            for lookup in self._variable_lookup:
                if node.id in lookup:
                    is_builtin = False
                    break

            if not is_builtin and _representable(value=value):
                text = self._atok.get_text(node)
                self.reprs[text] = value

        self.generic_visit(node=node)

    def visit_Attribute(self, node: ast.Attribute) -> None:
        """Represent the attribute by dumping its source code."""
        if node in self._recomputed_values:
            value = self._recomputed_values[node]

            if _representable(value=value):
                text = self._atok.get_text(node)
                self.reprs[text] = value

        self.generic_visit(node=node)

    if sys.version_info >= (3, 8):

        def visit_NamedExpr(self, node: ast.NamedExpr) -> Any:
            """Represent the target with the value of the node."""
            if node in self._recomputed_values:
                value = self._recomputed_values[node]

                target = node.target

                if _representable(value=value):
                    self.reprs[target.id] = value

            self.generic_visit(node=node)

    def visit_Call(self, node: ast.Call) -> None:
        """Represent the call by dumping its source code."""
        if node in self._recomputed_values:
            value = self._recomputed_values[node]
            text = self._atok.get_text(node)

            self.reprs[text] = value

        self.generic_visit(node=node)

    def visit_ListComp(self, node: ast.ListComp) -> None:
        """Represent the list comprehension by dumping its source code."""
        if node in self._recomputed_values:
            value = self._recomputed_values[node]
            text = self._atok.get_text(node)

            self.reprs[text] = value

        self.generic_visit(node=node)

    def visit_SetComp(self, node: ast.SetComp) -> None:
        """Represent the set comprehension by dumping its source code."""
        if node in self._recomputed_values:
            value = self._recomputed_values[node]
            text = self._atok.get_text(node)

            self.reprs[text] = value

        self.generic_visit(node=node)

    def visit_DictComp(self, node: ast.DictComp) -> None:
        """Represent the dictionary comprehension by dumping its source code."""
        if node in self._recomputed_values:
            value = self._recomputed_values[node]
            text = self._atok.get_text(node)

            self.reprs[text] = value

        self.generic_visit(node=node)

    def visit_Subscript(self, node: ast.Subscript) -> None:
        """Represent the subscript with its source code."""
        if node in self._recomputed_values:
            value = self._recomputed_values[node]
            text = self._atok.get_text(node)

            self.reprs[text] = value

        self.generic_visit(node=node)


def is_lambda(a_function: AnyCallable) -> bool:
    """
    Check whether the function is a lambda function.

    >>> def some_func()->bool: return True
    >>> is_lambda(some_func)
    False

    >>> lmbd = lambda x: x > 0
    >>> is_lambda(lmbd)
    True

    :return: True if condition is defined as lambda function
    """
    return a_function.__name__ == "<lambda>"


class ConditionLambdaInspection:
    """Represent the inspection of the condition function given as a lambda."""

    def __init__(self, atok: asttokens.asttokens.ASTTokens, node: ast.Lambda) -> None:
        """
        Initialize.

        :param atok: parsed AST tree and tokens with added positional properties
        :param node: lambda AST node corresponding to the condition
        """
        self.atok = atok
        self.node = node

        text = atok.get_text(node.body)
        assert isinstance(text, str)
        self.text = text


_DECORATOR_RE = re.compile(r"^\s*@[a-zA-Z_]")
_DEF_CLASS_RE = re.compile(r"^\s*(async\s+def|def |class )")


class DecoratorInspection:
    """Represent the inspection of a decorator extracted from a source file and embedded in a dummy dynamic module."""

    def __init__(self, atok: asttokens.asttokens.ASTTokens, node: ast.Call) -> None:
        """
        Initialize.

        :param atok: parsed AST tree and tokens with added positional properties
        :param node: lambda AST node corresponding to the condition
        """
        self.atok = atok
        self.node = node


def inspect_decorator(
    lines: list[str], lineno: int, filename: str
) -> DecoratorInspection:
    """
    Parse the file in which the decorator is called and figure out the corresponding call AST node.

    :param lines: lines of the source file corresponding to the decorator call
    :param lineno: line index (starting with 0) of one of the lines in the decorator call
    :param filename: name of the file where decorator is called
    :return: inspected decorator call
    """
    if lineno < 0 or lineno >= len(lines):
        raise ValueError(
            (
                "Given line number {} of one of the decorator lines "
                "is not within the range [{}, {}) of lines in {}.\n\n"
                "The decorator lines were:\n{}"
            ).format(lineno, 0, len(lines), filename, "\n".join(lines))
        )

    # Go up till a line starts with a decorator
    decorator_lineno = None  # type: int | None
    for i in range(lineno, -1, -1):
        if _DECORATOR_RE.match(lines[i]):
            decorator_lineno = i
            break

    if decorator_lineno is None:
        raise SyntaxError(
            "Decorator corresponding to the line {} could not be found in file {}: {!r}".format(
                lineno + 1, filename, lines[lineno]
            )
        )

    # Find the decorator end -- it's either a function definition, a class definition or another decorator
    decorator_end_lineno = None  # type: int | None
    for i in range(lineno + 1, len(lines)):
        line = lines[i]

        if _DECORATOR_RE.match(line) or _DEF_CLASS_RE.match(line):
            decorator_end_lineno = i
            break

    if decorator_end_lineno is None:
        raise SyntaxError(
            (
                "The next statement following the decorator corresponding to the line {} "
                "could not be found in file {}: {!r}"
            ).format(lineno + 1, filename, lines[lineno])
        )

    decorator_lines = lines[decorator_lineno:decorator_end_lineno]

    # We need to dedent the decorator and add a dummy decorate so that we can parse its text as valid source code.
    decorator_text = textwrap.dedent(
        "".join(decorator_lines)
    ) + "def dummy_{}(): pass".format(uuid.uuid4().hex)

    atok = asttokens.asttokens.ASTTokens(decorator_text, parse=True)

    if not isinstance(atok.tree, ast.Module):
        raise ValueError(
            (
                "Expected the parsed decorator text to live in an AST module. "
                "Are you trying to inspect a condition lambda which was not stated in a decorator? "
                "(This feature is currently unsupported in icontract.) "
                "The decorator was expected at line {} in {}. "
                "The decorator lines under inspection were {}-{}."
            ).format(lineno + 1, filename, decorator_lineno + 1, decorator_end_lineno)
        )

    module_node = atok.tree

    if len(module_node.body) != 1:
        raise ValueError(
            (
                "Expected the module AST of the decorator text to have a single statement. "
                "Are you trying to inspect a condition lambda which was not stated in a decorator? "
                "(This feature is currently unsupported in icontract.) "
                "The decorator was expected at line {} in {}. "
                "The decorator lines under inspection were {}-{}."
            ).format(lineno + 1, filename, decorator_lineno + 1, decorator_end_lineno)
        )

    if not isinstance(module_node.body[0], ast.FunctionDef):
        raise ValueError(
            (
                "Expected the only statement in the AST module corresponding to the decorator text "
                "to be a function definition. "
                "Are you trying to inspect a condition lambda which was not stated in a decorator? "
                "(This feature is currently unsupported in icontract.) "
                "The decorator was expected at line {} in {}. "
                "The decorator lines under inspection were {}-{}."
            ).format(lineno + 1, filename, decorator_lineno + 1, decorator_end_lineno)
        )

    func_def_node = module_node.body[0]

    if len(func_def_node.decorator_list) != 1:
        raise ValueError(
            (
                "Expected the function AST node corresponding to the decorator text to have a single decorator. "
                "Are you trying to inspect a condition lambda which was not stated in a decorator? "
                "(This feature is currently unsupported in icontract.) "
                "The decorator was expected at line {} in {}. "
                "The decorator lines under inspection were {}-{}."
            ).format(lineno + 1, filename, decorator_lineno + 1, decorator_end_lineno)
        )

    if not isinstance(func_def_node.decorator_list[0], ast.Call):
        raise ValueError(
            (
                "Expected the only decorator in the function definition AST node corresponding "
                "to the decorator text to be a call node. "
                "Are you trying to inspect a condition lambda which was not stated in a decorator? "
                "(This feature is currently unsupported in icontract.) "
                "The decorator was expected at line {} in {}. "
                "The decorator lines under inspection were {}-{}."
            ).format(lineno + 1, filename, decorator_lineno + 1, decorator_end_lineno)
        )

    call_node = func_def_node.decorator_list[0]

    return DecoratorInspection(atok=atok, node=call_node)


def find_lambda_condition(
    decorator_inspection: DecoratorInspection,
) -> ConditionLambdaInspection | None:
    """
    Inspect the decorator and extract the condition as lambda.

    If the condition is not given as a lambda function, return None.
    """
    call_node = decorator_inspection.node

    lambda_node = None  # type: ast.Lambda | None

    if len(call_node.args) > 0:
        assert isinstance(call_node.args[0], ast.Lambda), (
            "Expected the first argument to the decorator to be a condition as lambda AST node, "
            "but got: {}"
        ).format(type(call_node.args[0]))

        lambda_node = call_node.args[0]

    elif len(call_node.keywords) > 0:
        for keyword in call_node.keywords:
            if keyword.arg == "condition":
                assert isinstance(keyword.value, ast.Lambda), (
                    "Expected lambda node as value of the 'condition' argument to the decorator."
                )

                lambda_node = keyword.value
                break

        assert lambda_node is not None, (
            "Expected to find a keyword AST node with 'condition' arg, but found none"
        )
    else:
        raise AssertionError(
            "Expected a call AST node of a decorator to have either args or keywords, but got: {}".format(
                ast.dump(call_node)
            )
        )

    return ConditionLambdaInspection(atok=decorator_inspection.atok, node=lambda_node)


def inspect_lambda_condition(
    condition: Callable[..., Any],
) -> ConditionLambdaInspection | None:
    """
    Try to extract the source code of the condition as lambda.

    If the condition is not a lambda, returns None.
    """
    if not is_lambda(condition):
        return None

    lines, condition_lineno = inspect.findsource(condition)
    filename = inspect.getsourcefile(condition)
    assert filename is not None

    decorator_inspection = inspect_decorator(
        lines=lines, lineno=condition_lineno, filename=filename
    )

    lambda_inspection = find_lambda_condition(decorator_inspection=decorator_inspection)

    return lambda_inspection


# fmt: off
def collect_variable_lookup(
        condition: Callable[..., Any],
        resolved_kwargs: Mapping[str, Any] | None = None
) -> list[Mapping[str, Any]]:
    """
    Collect the variable lookups in order of precedence.

    :param condition: contract condition for which we are constructing a variable look-up tables
    :param resolved_kwargs:
        keyword arguments of the original function to be passed over to the condition (where applicable).

        The keyword arguments are added to the variable look-up tables accordingly.
        If ``resolved_kwargs`` is None, no keyword arguments will be added to the variable look-up table.
    """
    # fmt: on
    variable_lookup = []  # type: list[Mapping[str, Any]]

    ##
    # Condition-specific kwargs
    ##

    if resolved_kwargs is not None:
        variable_lookup.append(resolved_kwargs)

    ##
    # Add closure to the lookup
    ##

    closure_dict = dict()  # type: dict[str, Any]

    if condition.__closure__ is not None:
        closure_cells = condition.__closure__
        freevars = condition.__code__.co_freevars

        assert len(closure_cells) == len(freevars), \
            "Number of closure cells of a condition function ({}) == number of free vars ({})".format(
                len(closure_cells), len(freevars))

        for cell, freevar in zip(closure_cells, freevars):
            closure_dict[freevar] = cell.cell_contents

    variable_lookup.append(closure_dict)

    ##
    # Add globals to the lookup
    ##

    if condition.__globals__ is not None:
        variable_lookup.append(condition.__globals__)

    return variable_lookup


def repr_values(condition: Callable[..., bool], lambda_inspection: ConditionLambdaInspection | None,
                resolved_kwargs: Mapping[str, Any], a_repr: reprlib.Repr) -> list[str]:
    """
    Represent function arguments and frame values in the error message on contract breach.

    :param condition: condition function of the contract
    :param lambda_inspection:
        inspected lambda AST node corresponding to the condition function (None if the condition was not given as a
        lambda function)
    :param resolved_kwargs: arguments put in the function call
    :param a_repr: representation instance that defines how the values are represented.
    :return: list of value representations
    """
    # Hide _ARGS and _KWARGS if they are not part of the condition for better readability
    if '_ARGS' in resolved_kwargs or '_KWARGS' in resolved_kwargs:
        parameters = inspect.signature(condition).parameters
        malleable_kwargs = cast(
            MutableMapping[str, Any],
            resolved_kwargs.copy()  # type: ignore
        )

        if '_ARGS' not in parameters:
            malleable_kwargs.pop('_ARGS', None)

        if '_KWARGS' not in parameters:
            malleable_kwargs.pop('_KWARGS', None)

        selected_kwargs = cast(Mapping[str, Any], malleable_kwargs)
    else:
        selected_kwargs = resolved_kwargs

    # Don't use ``resolved_kwargs`` from this point on.
    # ``selected_kwargs`` is meant to be used instead for better readability of error messages.

    if is_lambda(a_function=condition):
        assert lambda_inspection is not None, "Expected a lambda inspection when given a condition as a lambda function"
    else:
        assert lambda_inspection is None, "Expected no lambda inspection in a condition given as a non-lambda function"

    reprs = None  # type: MutableMapping[str, Any] | None

    if lambda_inspection is not None:
        variable_lookup = collect_variable_lookup(condition=condition, resolved_kwargs=selected_kwargs)

        recompute_visitor = _recompute.Visitor(variable_lookup=variable_lookup)

        recompute_visitor.visit(node=lambda_inspection.node.body)
        recomputed_values = recompute_visitor.recomputed_values

        repr_visitor = Visitor(
            recomputed_values=recomputed_values, variable_lookup=variable_lookup, atok=lambda_inspection.atok)
        repr_visitor.visit(node=lambda_inspection.node.body)

        reprs = repr_visitor.reprs

    # Add original arguments from the call unless they shadow a variable in the re-computation.
    #
    # The condition arguments are often not sufficient to figure out the error. The user usually needs
    # more context which is captured in the remainder of the call arguments.

    if reprs is None:
        reprs = dict()

    for key in sorted(selected_kwargs.keys()):
        val = selected_kwargs[key]
        if key not in reprs and _representable(value=val):
            reprs[key] = val

    parts = []  # type: list[str]

    # We need to sort in order to present the same violation error on repeated violations.
    # Otherwise, the order of the reported arguments may be arbitrary.
    for key in sorted(reprs.keys()):
        value = reprs[key]
        if isinstance(value, _recompute.FirstExceptionInAll):
            writing = ['{} was False, e.g., with'.format(key)]
            for input_name, input_value in value.inputs:
                writing.append('\n')
                writing.append('  {} = {}'.format(input_name, a_repr.repr(input_value)))

            parts.append(''.join(writing))
        else:
            parts.append('{} was {}'.format(key, a_repr.repr(value)))

    return parts


def represent_condition(condition: AnyCallable) -> str:
    """Represent the condition as a string."""
    lambda_inspection = None  # type: ConditionLambdaInspection | None
    if not is_lambda(a_function=condition):
        condition_repr = condition.__name__
    else:
        # We need to extract the source code corresponding to the decorator since inspect.getsource() is broken with
        # lambdas.
        lambda_inspection = inspect_lambda_condition(condition=condition)
        assert lambda_inspection is not None, "Unexpected no lambda inspection for condition: {}".format(condition)
        condition_repr = lambda_inspection.atok.get_text(lambda_inspection.node)

    return condition_repr


def generate_message(contract: Contract, resolved_kwargs: Mapping[str, Any]) -> str:
    """Generate the message upon contract violation."""
    parts = []  # type: list[str]

    if contract.location is not None:
        parts.append("{}:\n".format(contract.location))

    if contract.description is not None:
        parts.append("{}: ".format(contract.description))

    lambda_inspection = None  # type: ConditionLambdaInspection | None
    if not is_lambda(a_function=contract.condition):
        condition_text = contract.condition.__name__
    else:
        # We need to extract the source code corresponding to the decorator since inspect.getsource() is broken with
        # lambdas.
        lambda_inspection = inspect_lambda_condition(condition=contract.condition)
        assert lambda_inspection is not None, \
            "Unexpected no lambda inspection for condition: {}".format(contract.condition)
        condition_text = lambda_inspection.text

    parts.append(condition_text)

    repr_vals = repr_values(
        condition=contract.condition,
        lambda_inspection=lambda_inspection,
        resolved_kwargs=resolved_kwargs,
        a_repr=contract._a_repr)

    if len(repr_vals) == 0:
        # Do not append anything since no value could be represented as a string.
        # This could appear in case we have, for example, a generator expression as the return value of a lambda.
        pass

    elif len(repr_vals) == 1:
        parts.append(': ')
        parts.append(repr_vals[0])
    else:
        parts.append(':\n')
        parts.append('\n'.join(repr_vals))

    msg = "".join(parts)

    return msg



from __future__ import annotations

import sys

# Before we do anything else, check Python >= 3.10 or raise an error.
if sys.version_info < (3, 10):
    raise RuntimeError(
        "nshutils.typecheck requires Python 3.10 or higher. "
        "Please upgrade your Python version to use this module."
    )

import logging
import os
import typing
from collections.abc import Callable, Sequence
from types import FrameType as _FrameType
from typing import Any, TypeAlias, Union

import wadler_lindig as wl
from beartype import beartype
from jaxtyping import BFloat16 as BFloat16
from jaxtyping import Bool as Bool
from jaxtyping import Complex as Complex
from jaxtyping import Complex64 as Complex64
from jaxtyping import Complex128 as Complex128
from jaxtyping import Float as Float
from jaxtyping import Float16 as Float16
from jaxtyping import Float32 as Float32
from jaxtyping import Float64 as Float64
from jaxtyping import Inexact as Inexact
from jaxtyping import Int as Int
from jaxtyping import Int4 as Int4
from jaxtyping import Int8 as Int8
from jaxtyping import Int16 as Int16
from jaxtyping import Int32 as Int32
from jaxtyping import Int64 as Int64
from jaxtyping import Integer as Integer
from jaxtyping import Key as Key
from jaxtyping import Num as Num
from jaxtyping import Real as Real
from jaxtyping import Shaped as Shaped
from jaxtyping import UInt as UInt
from jaxtyping import UInt4 as UInt4
from jaxtyping import UInt8 as UInt8
from jaxtyping import UInt16 as UInt16
from jaxtyping import UInt32 as UInt32
from jaxtyping import UInt64 as UInt64
from jaxtyping import jaxtyped
from jaxtyping._storage import get_shape_memo, shape_str
from typing_extensions import TypeVar

from ._tree_util import set_pytree_backend as set_pytree_backend

if typing.TYPE_CHECKING:
    # Hack taken directly from `jaxtyping`.
    # Set up to deliberately confuse a static type checker.
    PyTree: TypeAlias = getattr(typing, "foo" + "bar")  # pyright: ignore[reportInvalidTypeForm]
    # What's going on with this madness?
    #
    # At static-type-checking-time, we want `PyTree` to be a type for which both
    # `PyTree` and `PyTree[Foo]` are equivalent to `Any`.
    # (The intention is that `PyTree` be a runtime-only type; there's no real way to
    # do more with static type checkers.)
    #
    # Unfortunately, this isn't possible: `Any` isn't subscriptable. And there's no
    # equivalent way we can fake this using typing annotations. (In some sense the
    # closest thing would be a `Protocol[T]` with no methods, but that's actually the
    # opposite of what we want: that ends up allowing nothing at all.)
    #
    # The good news for us is that static type checkers have an internal escape hatch.
    # If they can't figure out what a type is, then they just give up and allow
    # anything. (I believe this is sometimes called `Unknown`.) Thus, this odd-looking
    # annotation, which static type checkers aren't smart enough to resolve.
else:
    from ._pytree_type_dynamic import PyTree as PyTree

log = logging.getLogger(__name__)

DISABLE_ENV_KEY = "NSHUTILS_DISABLE_TYPECHECKING"


def typecheck_modules(modules: Sequence[str]):
    """
    Typecheck the given modules using `jaxtyping`.

    Args:
        modules: Modules to typecheck.
    """
    # If `DISABLE_ENV_KEY` is set and the environment variable is set, skip
    #   typechecking.
    if DISABLE_ENV_KEY is not None and bool(int(os.environ.get(DISABLE_ENV_KEY, "0"))):
        log.critical(
            f"Type checking is disabled due to the environment variable {DISABLE_ENV_KEY}."
        )
        return

    # Install the jaxtyping import hook for this module.
    from jaxtyping import install_import_hook

    install_import_hook(modules, "beartype.beartype")

    log.critical(f"Type checking the following modules: {modules}")


def _get_frame_package_name_or_none(frame: _FrameType) -> str | None:
    # Taken from `beartype._util.func.utilfuncframe.get_frame_package_name_or_none`.
    assert isinstance(frame, _FrameType), f"{repr(frame)} not stack frame."

    # Fully-qualified name of the parent package of the child module declaring
    # the callable whose code object is that of this stack frame's if that
    # module declares its name *OR* the empty string otherwise (e.g., if that
    # module is either a top-level module or script residing outside any parent
    # package structure).
    frame_package_name = frame.f_globals.get("__package__")

    # Return the name of this parent package.
    return frame_package_name


def typecheck_this_module(additional_modules: Sequence[str] = ()):
    """
    Typecheck the calling module and any additional modules using `jaxtyping`.

    Args:
        additional_modules: Additional modules to typecheck.
    """
    # Get the calling module's name.
    # Here, we can just use beartype's internal implementation behind
    # `beartype_this_package`.

    # Get the calling module's name.
    frame = sys._getframe(1)
    assert frame is not None, "frame is None"
    calling_module_name = _get_frame_package_name_or_none(frame)
    assert calling_module_name is not None, "calling_module_name is None"

    # Typecheck the calling module + any additional modules.
    typecheck_modules((calling_module_name, *additional_modules))


def _make_error_str(input: Any, t: Any) -> str:
    error_components: list[str] = []
    error_components.append("Type checking error:")
    if hasattr(t, "__instancecheck_str__"):
        error_components.append(t.__instancecheck_str__(input))

    error_components.append(wl.pformat(input))
    error_components.append(shape_str(get_shape_memo()))

    return "\n".join(error_components)


def tassert(t: Any, input: object):
    """
    Typecheck the input against the given type.

    Args:
        t: Type to check against.
        input: Input to check.
    """
    __tracebackhide__ = True

    # Ignore typechecking if the environment variable is set.
    if DISABLE_ENV_KEY is not None and bool(int(os.environ.get(DISABLE_ENV_KEY, "0"))):
        return

    assert isinstance(input, t), _make_error_str(input, t)


_TypeOrCallable = TypeVar(
    "_TypeOrCallable",
    bound=Union[type, Callable],
)


def typecheck(fn: _TypeOrCallable, /) -> _TypeOrCallable:
    """
    Decorator to typecheck the function's arguments and return value.
    This decorator uses `jaxtyping` to enforce type checking.
    Args:
        fn: Function to typecheck.
    Returns:
        The typechecked function.
    """

    __tracebackhide__ = True

    # Ignore typechecking if the environment variable is set.
    if DISABLE_ENV_KEY is not None and bool(int(os.environ.get(DISABLE_ENV_KEY, "0"))):
        return fn

    return jaxtyped(typechecker=beartype)(fn)
